"""scrapli.exceptions"""


class ScrapliTimeout(Exception):
    """Exception for any scrapli timeouts"""


class ScrapliKeepaliveFailure(Exception):
    """Exception for scrapli missing keepalives"""


class MissingDependencies(Exception):
    """Exception for any missing (probably optional) dependencies"""


class KeyVerificationFailed(Exception):
    """Exception for scrapli public key verification failure"""


class ScrapliAuthenticationFailed(Exception):
    """Exception for scrapli authentication failure"""


class UnknownPrivLevel(Exception):
    """Exception for encountering unknown privilege level"""


class CouldNotAcquirePrivLevel(Exception):
    """Exception for failure to acquire desired privilege level"""



"""scrapli.decorators"""
import logging
from concurrent.futures import ThreadPoolExecutor, wait
from typing import TYPE_CHECKING, Any, Callable, Dict, Union

if TYPE_CHECKING:
    from scrapli.channel import Channel
    from scrapli.transport import Transport

LOG = logging.getLogger("scrapli")


def operation_timeout(attribute: str, message: str = "") -> Callable[..., Any]:
    """
    Decorate an "operation" -- raises exception if the operation timeout is exceeded

    Wrap an operation, check class for given attribute and use that for the timeout duration.

    Historically this operation timeout decorator used signals instead of the concurrent_futures
    seen here. The signals method was probably a bit more elegant, however there were issues with
    supporting the system transport as system transport subprocess/ptyprocess components spawn
    threads of their own, and signals must operate in the main thread.

    Args:
        attribute: attribute to inspect in class (to set timeout duration)
        message: optional message to pass when decorating a non-standard operation such as telnet
            login (as opposed to "normal" channel operations)

    Returns:
        decorate: wrapped function

    Raises:
        TimeoutError: if timeout exceeded

    """

    def decorate(wrapped_func: Callable[..., Any]) -> Callable[..., Any]:
        def timeout_wrapper(
            channel_or_transport: Union["Channel", "Transport"],
            *args: Union[str, int],
            **kwargs: Dict[str, Union[str, int]],
        ) -> Any:
            # import here to avoid circular dependency
            from scrapli.channel import Channel  # pylint: disable=C0415

            timeout_duration = getattr(channel_or_transport, attribute, None)
            if not timeout_duration:
                LOG.error(
                    f"Could not find {attribute} value of {channel_or_transport}, continuing "
                    "without timeout decorator"
                )
                return wrapped_func(channel_or_transport, *args, **kwargs)

            # as this can be called from transport or channel get the appropriate objects
            # to unlock and close the session. we need to unlock as the close will block
            # forever if the session is locked, and the session very likely is locked while
            # waiting for output from the device
            if isinstance(channel_or_transport, Channel):
                timeout_exit = channel_or_transport.transport.timeout_exit
                session_lock = channel_or_transport.transport.session_lock
                close = channel_or_transport.transport.close
            else:
                timeout_exit = channel_or_transport.timeout_exit
                session_lock = channel_or_transport.session_lock
                close = channel_or_transport.close

            pool = ThreadPoolExecutor(max_workers=1)
            future = pool.submit(wrapped_func, channel_or_transport, *args, **kwargs)
            wait([future], timeout=timeout_duration)
            if not future.done():
                LOG.info(message)
                if timeout_exit:
                    LOG.info("timeout_exit is True, closing transport")
                    if session_lock.locked():
                        session_lock.release()
                    close()
                raise TimeoutError(message)
            return future.result()

        return timeout_wrapper

    return decorate



"""scrapli.transport.cssh2"""
import base64
import time
import warnings
from logging import getLogger
from typing import Optional, Tuple

from scrapli.exceptions import (
    KeyVerificationFailed,
    MissingDependencies,
    ScrapliAuthenticationFailed,
)
from scrapli.ssh_config import SSHConfig, SSHKnownHosts
from scrapli.transport.socket import Socket
from scrapli.transport.transport import Transport

LOG = getLogger("transport")


SSH2_TRANSPORT_ARGS = (
    "auth_username",
    "auth_private_key",
    "auth_password",
    "auth_strict_key",
    "ssh_config_file",
    "ssh_known_hosts_file",
    "timeout_socket",
)


class SSH2Transport(Transport):
    def __init__(
        self,
        host: str,
        port: int = -1,
        auth_username: str = "",
        auth_private_key: str = "",
        auth_password: str = "",
        auth_strict_key: bool = True,
        timeout_socket: int = 5,
        timeout_transport: int = 5,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "",
        keepalive_pattern: str = "\005",
        ssh_config_file: str = "",
        ssh_known_hosts_file: str = "",
    ) -> None:
        """
        SSH2Transport Object

        Inherit from Transport ABC
        SSH2Transport <- Transport (ABC)

        Args:
            host: host ip/name to connect to
            port: port to connect to
            auth_username: username for authentication
            auth_private_key: path to private key for authentication
            auth_password: password for authentication
            auth_strict_key: True/False to enforce strict key checking (default is True)
            timeout_socket: timeout for establishing socket in seconds
            timeout_transport: timeout for ssh2 transport in seconds
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                'standard' keepalive mechanisms. 'standard' attempts to ssh2-python built in
                keepalive method (using standard openssh keepalive)
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'
            ssh_config_file: string to path for ssh config file
            ssh_known_hosts_file: string to path for ssh known hosts file

        Returns:
            N/A  # noqa: DAR202

        Raises:
            MissingDependencies: if ssh2-python is not installed

        """
        cfg_port, cfg_user, cfg_private_key = self._process_ssh_config(host, ssh_config_file)

        if port == -1:
            port = cfg_port or 22

        super().__init__(
            host,
            port,
            timeout_socket,
            timeout_transport,
            timeout_exit,
            keepalive,
            keepalive_interval,
            keepalive_type,
            keepalive_pattern,
        )

        self.auth_username: str = auth_username or cfg_user
        self.auth_private_key: str = auth_private_key or cfg_private_key
        self.auth_password: str = auth_password
        self.auth_strict_key: bool = auth_strict_key
        self.ssh_known_hosts_file: str = ssh_known_hosts_file

        try:
            # import here so these are optional
            from ssh2.channel import Channel  # pylint: disable=C0415
            from ssh2.session import Session  # pylint: disable=C0415
            from ssh2.exceptions import AuthenticationError, SSH2Error  # pylint: disable=C0415

            self.lib_session = Session
            self.session: Session = None
            self.channel: Channel = None
            self.lib_auth_exception = AuthenticationError
            self.lib_base_exception = SSH2Error
        except ModuleNotFoundError as exc:
            err = f"Module '{exc.name}' not installed!"
            msg = f"***** {err} {'*' * (80 - len(err))}"
            fix = (
                f"To resolve this issue, install '{exc.name}'. You can do this in one of the "
                "following ways:\n"
                "1: 'pip install -r requirements-ssh2.txt'\n"
                "2: 'pip install scrapli[ssh2]'"
            )
            warning = "\n" + msg + "\n" + fix + "\n" + msg
            warnings.warn(warning)
            LOG.warning(warning)
            raise MissingDependencies

        self.socket = Socket(host=self.host, port=self.port, timeout=self.timeout_socket)

    @staticmethod
    def _process_ssh_config(host: str, ssh_config_file: str) -> Tuple[Optional[int], str, str]:
        """
        Method to parse ssh config file

        In the future this may move to be a "helper" function as it should be very similar between
        paramiko and and ssh2-python... for now it can be a static method as there may be varying
        supported args between the two transport drivers.

        Args:
            host: host to lookup in ssh config file
            ssh_config_file: string path to ssh config file; passed down from `Scrape`, or the
                `NetworkDriver` or subclasses of it, in most cases.

        Returns:
            Tuple: port to use for ssh, username to use for ssh, identity file (private key) to
                use for ssh auth

        Raises:
            N/A

        """
        ssh = SSHConfig(ssh_config_file)
        host_config = ssh.lookup(host)
        return host_config.port, host_config.user or "", host_config.identity_file or ""

    def open(self) -> None:
        """
        Parent method to open session, authenticate and acquire shell

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            exc: if socket handshake fails
            ScrapliAuthenticationFailed: if all authentication means fail

        """
        if not self.socket.socket_isalive():
            self.socket.socket_open()
        self.session_lock.acquire()
        self.session = self.lib_session()
        self.set_timeout(self.timeout_transport)
        try:
            self.session.handshake(self.socket.sock)
        except Exception as exc:
            LOG.critical(
                f"Failed to complete handshake with host {self.host}; " f"Exception: {exc}"
            )
            self.session_lock.release()
            raise exc

        if self.auth_strict_key:
            LOG.debug(f"Attempting to validate {self.host} public key")
            self._verify_key()

        LOG.debug(f"Session to host {self.host} opened")
        self.authenticate()
        if not self.isauthenticated():
            msg = f"Authentication to host {self.host} failed"
            LOG.critical(msg)
            self.session_lock.release()
            raise ScrapliAuthenticationFailed(msg)
        self._open_channel()
        if self.keepalive:
            self._session_keepalive()
        self.session_lock.release()

    def _verify_key(self) -> None:
        """
        Verify target host public key, raise exception if invalid/unknown

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            KeyVerificationFailed: if public key verification fails

        """
        known_hosts = SSHKnownHosts(self.ssh_known_hosts_file)

        if self.host not in known_hosts.hosts.keys():
            raise KeyVerificationFailed(f"{self.host} not in known_hosts!")

        remote_server_key_info = self.session.hostkey()
        encoded_remote_server_key = remote_server_key_info[0]
        raw_remote_public_key = base64.encodebytes(encoded_remote_server_key)
        remote_public_key = raw_remote_public_key.replace(b"\n", b"").decode()

        if known_hosts.hosts[self.host]["public_key"] != remote_public_key:
            raise KeyVerificationFailed(
                f"{self.host} in known_hosts but public key does not match!"
            )

    def authenticate(self) -> None:
        """
        Parent method to try all means of authentication

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if self.auth_private_key:
            self._authenticate_public_key()
            if self.isauthenticated():
                LOG.debug(f"Authenticated to host {self.host} with public key auth")
                return
            if not self.auth_password or not self.auth_username:
                msg = (
                    f"Public key authentication to host {self.host} failed. Missing username or"
                    " password unable to attempt password authentication."
                )
                LOG.critical(msg)
                raise ScrapliAuthenticationFailed(msg)

        self._authenticate_password()
        if self.isauthenticated():
            LOG.debug(f"Authenticated to host {self.host} with password")
            return
        self._authenticate_keyboard_interactive()
        if self.isauthenticated():
            LOG.debug(f"Authenticated to host {self.host} with keyboard interactive")

    def _authenticate_public_key(self) -> None:
        """
        Attempt to authenticate with public key authentication

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            exc: if unknown (i.e. not auth failed) exception occurs

        """
        try:
            self.session.userauth_publickey_fromfile(self.auth_username, self.auth_private_key)
        except self.lib_auth_exception as exc:
            LOG.critical(
                f"Public key authentication with host {self.host} failed. Exception: {exc}."
            )
        except self.lib_base_exception as exc:
            LOG.critical(
                "Unknown error occurred during public key authentication with host "
                f"{self.host}; Exception: {exc}"
            )

    def _authenticate_password(self) -> None:
        """
        Attempt to authenticate with password authentication

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            exc: if unknown (i.e. not auth failed) exception occurs

        """
        try:
            self.session.userauth_password(self.auth_username, self.auth_password)
        except self.lib_auth_exception as exc:
            LOG.critical(f"Password authentication with host {self.host} failed. Exception: {exc}.")
        except Exception as exc:
            LOG.critical(
                "Unknown error occurred during password authentication with host "
                f"{self.host}; Exception: {exc}"
            )
            raise exc

    def _authenticate_keyboard_interactive(self) -> None:
        """
        Attempt to authenticate with keyboard interactive authentication

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            exc: if unknown (i.e. not auth failed) exception occurs

        """
        try:
            self.session.userauth_keyboardinteractive(  # pylint: disable=C0415
                self.auth_username, self.auth_password
            )
        except AttributeError as exc:
            LOG.critical(
                "Keyboard interactive authentication not supported in your ssh2-python version. "
                f"Exception: {exc}"
            )
        except self.lib_auth_exception as exc:
            LOG.critical(
                f"Keyboard interactive authentication with host {self.host} failed. "
                f"Exception: {exc}."
            )
        except Exception as exc:
            LOG.critical(
                "Unknown error occurred during keyboard interactive authentication with host "
                f"{self.host}; Exception: {exc}"
            )
            raise exc

    def isauthenticated(self) -> bool:
        """
        Check if session is authenticated

        Args:
            N/A

        Returns:
            bool: True if authenticated, else False

        Raises:
            N/A

        """
        authenticated: bool = self.session.userauth_authenticated()
        return authenticated

    def _open_channel(self) -> None:
        """
        Open channel, acquire pty, request interactive shell

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.channel = self.session.open_session()
        self.channel.pty()
        self.channel.shell()
        LOG.debug(f"Channel to host {self.host} opened")

    def close(self) -> None:
        """
        Close session and socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session_lock.acquire()
        self.channel.close()
        LOG.debug(f"Channel to host {self.host} closed")
        self.socket.socket_close()
        self.session_lock.release()

    def isalive(self) -> bool:
        """
        Check if socket is alive and session is authenticated

        Args:
            N/A

        Returns:
            bool: True if socket is alive and session authenticated, else False

        Raises:
            N/A

        """
        if self.socket.socket_isalive() and not self.channel.eof() and self.isauthenticated():
            return True
        return False

    def read(self) -> bytes:
        """
        Read data from the channel

        Args:
            N/A

        Returns:
            bytes: bytes output as read from channel

        Raises:
            N/A

        """
        output: bytes
        _, output = self.channel.read(65535)
        return output

    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.channel.write(channel_input)

    def set_timeout(self, timeout: Optional[int] = None) -> None:
        """
        Set session timeout

        Args:
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if isinstance(timeout, int):
            set_timeout = timeout
        else:
            set_timeout = self.timeout_transport
        # ssh2-python expects timeout in milliseconds
        self.session.set_timeout(set_timeout * 1000)

    def _keepalive_standard(self) -> None:
        """
        Send "out of band" (protocol level) keepalives to devices.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session.keepalive_config(want_reply=False, interval=self.keepalive_interval)
        while True:
            if not self.isalive():
                return
            LOG.debug("Sending 'standard' keepalive.")
            self.session.keepalive_send()
            time.sleep(self.keepalive_interval / 10)



"""scrapli.transport.transport"""
import time
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime
from logging import getLogger
from threading import Lock
from typing import Optional

from scrapli.exceptions import ScrapliKeepaliveFailure

LOG = getLogger("transport")


class Transport(ABC):
    def __init__(
        self,
        host: str = "",
        port: int = 22,
        timeout_socket: int = 5,
        timeout_transport: int = 5,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "network",
        keepalive_pattern: str = "\005",
    ) -> None:
        """
        Transport Base Object

        Args:
            host: host ip/name to connect to
            port: port to connect to
            timeout_socket: timeout for establishing socket in seconds
            timeout_transport: timeout for ssh|telnet transport in seconds
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                'standard' keepalive mechanisms. 'standard' attempts to use whatever 'standard'
                keepalive mechanisms are available in the selected transport mechanism. Check the
                transport documentation for details on what is supported and/or how it is
                implemented for any given transport driver
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.host: str = host
        self.port: int = port
        self.timeout_socket: int = timeout_socket
        self.timeout_transport: int = timeout_transport
        self.timeout_exit: bool = timeout_exit
        self.keepalive: bool = keepalive
        self.keepalive_interval: int = keepalive_interval
        self.keepalive_type: str = keepalive_type
        self.keepalive_pattern: str = keepalive_pattern

        self.session_lock: Lock = Lock()

    def __bool__(self) -> bool:
        """
        Magic bool method for Socket

        Args:
            N/A

        Returns:
            bool: True/False if socket is alive or not

        Raises:
            N/A

        """
        return self.isalive()

    def __str__(self) -> str:
        """
        Magic str method for Transport

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        return f"Transport Object for host {self.host}"

    def __repr__(self) -> str:
        """
        Magic repr method for Transport

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        class_dict["auth_password"] = "********"
        return f"Transport {class_dict}"

    @abstractmethod
    def open(self) -> None:
        """
        Open channel, acquire pty, request interactive shell

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def close(self) -> None:
        """
        Close session and socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def isalive(self) -> bool:
        """
        Check if socket is alive and session is authenticated

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def read(self) -> bytes:
        """
        Read data from the channel

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    @abstractmethod
    def set_timeout(self, timeout: Optional[int] = None) -> None:
        """
        Set session timeout

        Args:
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """

    def _session_keepalive(self) -> None:
        """
        Spawn keepalive thread for transport session

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if not self.keepalive:
            return
        pool = ThreadPoolExecutor()
        if self.keepalive_type == "network":
            pool.submit(self._keepalive_network)
        else:
            pool.submit(self._keepalive_standard)

    def _keepalive_network(self) -> None:
        """
        Send "in band" keepalives to devices.

        Generally used with "network" devices which do not have native keepalive support. This will
        try to acquire a session lock and send an innocuous character -- such as CTRL+E -- to keep
        the device "exec-timeout" (in network-y words) from expiring.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliKeepaliveFailure: if scrapli cant unlock and send keepalive in less than 3 *
                the keepalive_interval

        """
        lock_counter = 0
        last_keepalive = datetime.now()
        while True:
            if not self.isalive():
                return
            diff = datetime.now() - last_keepalive
            if diff.seconds >= self.keepalive_interval:
                if not self.session_lock.locked():
                    LOG.debug(
                        f"Sending 'network' keepalive with pattern {repr(self.keepalive_pattern)}."
                    )
                    lock_counter = 0
                    self.session_lock.acquire()
                    self.write(self.keepalive_pattern)
                    self.session_lock.release()
                    last_keepalive = datetime.now()
                else:
                    lock_counter += 1
                    if lock_counter >= 3:
                        LOG.info(f"Keepalive thread missed {lock_counter} consecutive keepalives.")
            if diff.seconds > self.keepalive_interval * 3:
                msg = (
                    "Keepalive thread has failed to send a keepalive in greater than three "
                    "times the keepalive interval!"
                )
                LOG.critical(msg)
                raise ScrapliKeepaliveFailure(msg)
            time.sleep(self.keepalive_interval / 10)

    @abstractmethod
    def _keepalive_standard(self) -> None:
        """
        Send "out of band" (protocol level) keepalives to devices.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """



import time

import pytest

from scrapli.exceptions import ScrapliAuthenticationFailed

from .test_data.devices import DEVICES, INVALID_PRIVATE_KEY, PRIVATE_KEY
from .test_data.test_cases import TEST_CASES


class TestDevice:
    def test_get_prompt(self, nix_conn, transport):
        prompt = nix_conn.channel.get_prompt()
        assert prompt == TEST_CASES["linux"]["get_prompt"]

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_short(self, nix_conn, transport, strip_prompt):
        command = TEST_CASES["linux"]["send_command_short"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES["linux"]["send_command_short"][expected_type]
        response = nix_conn.channel.send_input(channel_input=command, strip_prompt=strip_prompt)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_long(self, nix_conn, transport, strip_prompt):
        command = TEST_CASES["linux"]["send_command_long"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES["linux"]["send_command_long"][expected_type]
        response = nix_conn.channel.send_input(channel_input=command, strip_prompt=strip_prompt)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_commands(self, nix_conn, transport, strip_prompt):
        commands = []
        expected_responses = []
        commands.append(TEST_CASES["linux"]["send_command_short"]["command"])
        commands.append(TEST_CASES["linux"]["send_command_long"]["command"])
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses.append(TEST_CASES["linux"]["send_command_short"][expected_type])
        expected_responses.append(TEST_CASES["linux"]["send_command_long"][expected_type])
        responses = nix_conn.channel.send_inputs(channel_inputs=commands, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response

    def test_isalive_and_close(self, nix_conn, transport):
        assert nix_conn.isalive() is True
        nix_conn.close()
        # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
        # close in time for the next assert
        time.sleep(0.1)
        assert nix_conn.isalive() is False


class TestNetworkDevice:
    @pytest.mark.parametrize(
        "priv_level",
        ["exec", "privilege_exec", "configuration"],
        ids=["exec", "privilege_exec", "configuration"],
    )
    def test_get_prompt_and_acquire_priv(self, conn, priv_level, device_type, transport):
        if TEST_CASES[device_type]["get_prompt"][priv_level] is None:
            pytest.skip(f"Priv level {priv_level} for device type {device_type} not tested")
        conn.acquire_priv(priv_level)
        prompt = conn.get_prompt()
        assert prompt == TEST_CASES[device_type]["get_prompt"][priv_level]

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_short(self, conn, device_type, transport, strip_prompt):
        command = TEST_CASES[device_type]["send_command_short"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_command_short"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = conn.send_command(command=command, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_command_long(self, conn, device_type, transport, strip_prompt):
        command = TEST_CASES[device_type]["send_command_long"]["command"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_response = TEST_CASES[device_type]["send_command_long"][expected_type]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        response = conn.send_command(command=command, strip_prompt=strip_prompt)
        assert sanitize_response(response.result) == expected_response

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_commands(self, conn, device_type, transport, strip_prompt):
        commands = []
        expected_responses = []
        commands.append(TEST_CASES[device_type]["send_command_short"]["command"])
        commands.append(TEST_CASES[device_type]["send_command_long"]["command"])
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses.append(TEST_CASES[device_type]["send_command_short"][expected_type])
        expected_responses.append(TEST_CASES[device_type]["send_command_long"][expected_type])
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = conn.send_commands(commands=commands, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert sanitize_response(response.result) == expected_response

    def test_send_interactive_normal_response(self, conn, device_type, transport):
        if TEST_CASES[device_type]["send_interactive_normal_response"] is None:
            pytest.skip(
                f"send_interactive_normal_response for device type {device_type} not tested"
            )
        interact = TEST_CASES[device_type]["send_interactive_normal_response"]["command"]
        expected_response = TEST_CASES[device_type]["send_interactive_normal_response"]["expected"]
        response = conn.send_interactive(interact=interact)
        assert response.result == expected_response

    def test_send_interactive_hidden_response(self, conn, device_type, transport):
        if TEST_CASES[device_type]["send_interactive_hidden_response"] is None:
            pytest.skip(
                f"send_interactive_hidden_response for device type {device_type} not tested"
            )
        interact = TEST_CASES[device_type]["send_interactive_hidden_response"]["command"]
        expected_response = TEST_CASES[device_type]["send_interactive_hidden_response"]["expected"]
        response = conn.send_interactive(interact=interact)
        assert response.result == expected_response

    @pytest.mark.parametrize(
        "strip_prompt", [True, False], ids=["strip_prompt", "no_strip_prompt"],
    )
    def test_send_configs(self, conn, device_type, transport, strip_prompt):
        configs = TEST_CASES[device_type]["send_configs"]["configs"]
        expected_type = "expected_no_strip" if not strip_prompt else "expected_strip"
        expected_responses = TEST_CASES[device_type]["send_configs"][expected_type]
        verification = TEST_CASES[device_type]["send_configs"]["verification"]
        expected_verification = TEST_CASES[device_type]["send_configs"][
            f"verification_{expected_type}"
        ]
        teardown_configs = TEST_CASES[device_type]["send_configs"]["teardown_configs"]
        sanitize_response = TEST_CASES[device_type]["sanitize_response"]
        responses = conn.send_configs(configs=configs, strip_prompt=strip_prompt)
        for expected_response, response in zip(expected_responses, responses):
            assert response.result == expected_response
        verification_response = conn.send_command(command=verification, strip_prompt=strip_prompt)
        assert sanitize_response(verification_response.result) == expected_verification
        conn.send_configs(configs=teardown_configs)

    def test_isalive_and_close(self, conn, device_type, transport):
        assert conn.isalive() is True
        conn.close()
        # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
        # close in time for the next assert
        time.sleep(0.1)
        assert conn.isalive() is False


def test_context_manager(device_type, transport):
    if device_type == "arista_eos" and transport == "ssh2":
        pytest.skip(
            "SSH2 (on pypi) doesn't support keyboard interactive auth, skipping ssh2 for arista_eos testing"
        )

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")

    port = 22
    if transport == "telnet":
        port = 23

    device["port"] = port
    device["transport"] = transport
    device["timeout_socket"] = 2
    device["timeout_transport"] = 2
    device["timeout_ops"] = 2

    with driver(**device) as conn:
        assert conn.isalive() is True
    # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
    # close in time for the next assert
    time.sleep(0.1)
    assert conn.isalive() is False


def test_public_key_auth(device_type, transport):
    if device_type != "cisco_iosxe" or transport == "telnet":
        pytest.skip("public key auth only tested against iosxe at the moment, and never on telnet!")

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")

    device["transport"] = transport
    device["timeout_socket"] = 2
    device["timeout_transport"] = 2
    device["timeout_ops"] = 2
    device["auth_private_key"] = PRIVATE_KEY

    with driver(**device) as conn:
        assert conn.isalive() is True
    # unsure why but w/out a tiny sleep pytest just plows ahead and the connection doesnt
    # close in time for the next assert
    time.sleep(0.1)
    assert conn.isalive() is False


def test_public_key_auth_failure(device_type, transport):
    if device_type != "cisco_iosxe" or transport == "telnet":
        pytest.skip("public key auth only tested against iosxe at the moment, and never on telnet!")

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")

    device["transport"] = transport
    device["timeout_socket"] = 2
    device["timeout_transport"] = 2
    device["timeout_ops"] = 2
    device["auth_private_key"] = INVALID_PRIVATE_KEY
    device.pop("auth_password")
    conn = driver(**device)

    with pytest.raises(ScrapliAuthenticationFailed) as exc:
        conn.open()
    assert str(exc.value) == (
        f"Public key authentication to host {device['host']} failed. Missing username or "
        "password unable to attempt password authentication."
    )



"""scrapli.response"""
from datetime import datetime
from io import TextIOWrapper
from typing import Any, Dict, List, Optional, Union

from scrapli.helper import _textfsm_get_template, textfsm_parse


class Response:
    def __init__(
        self,
        host: str,
        channel_input: str,
        textfsm_platform: str = "",
        expectation: Optional[str] = None,
        channel_response: Optional[str] = None,
        finale: Optional[str] = None,
        failed_when_contains: Optional[List[str]] = None,
    ):
        """
        Scrapli Response

        Store channel_input, resulting output, and start/end/elapsed time information. Attempt to
        determine if command was successful or not and reflect that in a failed attribute.

        Args:
            host: host that was operated on
            channel_input: input that got sent down the channel
            textfsm_platform: ntc-templates friendly platform type
            expectation: used for send_inputs_interact -- string to expect back from the channel
                after initial input
            channel_response: used for send_inputs_interact -- string to use to respond to expected
                prompt
            finale: string of prompt to look for to know when "done" with interaction
            failed_when_contains: list of strings that, if present in final output, represent a
                failed command/interaction

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A  # noqa

        """
        self.host = host
        self.start_time = datetime.now()
        self.finish_time: Optional[datetime] = None
        self.elapsed_time: Optional[float] = None

        self.channel_input = channel_input
        self.textfsm_platform = textfsm_platform
        self.expectation = expectation
        self.channel_response = channel_response
        self.finale = finale
        self.raw_result: str = ""
        self.result: str = ""

        # for future use -- could add failed when terms in each driver, then check for those strings
        # in results to determine if the command failed, could also set this at send_inputs level
        self.failed_when_contains = failed_when_contains
        self.failed = True

    def __bool__(self) -> bool:
        """
        Magic bool method based on channel_input being failed or not

        Args:
            N/A

        Returns:
            bool: True/False if channel_input failed

        Raises:
            N/A

        """
        return self.failed

    def __repr__(self) -> str:
        """
        Magic repr method for SSH2NetResponse class

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        return f"Scrape <Success: {str(not self.failed)}>"

    def __str__(self) -> str:
        """
        Magic str method for SSH2NetResponse class

        Args:
            N/A

        Returns:
            str: str for class object

        Raises:
            N/A

        """
        return f"Scrape <Success: {str(not self.failed)}>"

    def record_response(self, result: str) -> None:
        """
        Record channel_input results and elapsed time of channel input/reading output

        Args:
            result: string result of channel_input

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.finish_time = datetime.now()
        self.elapsed_time = (self.finish_time - self.start_time).total_seconds()
        self.result = result
        # update failed to false after recording results
        self.failed = False

    def textfsm_parse_output(self) -> Union[Dict[str, Any], List[Any]]:
        """
        Parse results with textfsm, assign result to `structured_result`

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        template = _textfsm_get_template(self.textfsm_platform, self.channel_input)
        if isinstance(template, TextIOWrapper):
            structured_result = textfsm_parse(template, self.result) or []
        else:
            structured_result = []
        return structured_result



import logging
import time
from pathlib import Path

from device_info import eos_device
from scrapli.driver.core import EOSDriver

logging.basicConfig(
    filename=f"{Path(__file__).resolve().parents[0]}/eos_driver.log", level=logging.DEBUG
)
logger = logging.getLogger("scrapli")

args = {
    "host": eos_device["host"],
    "port": eos_device["port"],
    "auth_username": eos_device["auth_username"],
    "auth_password": eos_device["auth_password"],
    "auth_strict_key": False,
    "keepalive_interval": eos_device["keepalive_interval"],
    "transport": eos_device["transport"],
    "keepalive": eos_device["keepalive"],
    "comms_ansi": True,
}

conn = EOSDriver(**args)
conn.open()

# vrnetlab eos has no enable password so we just enable this manually for now
# conn.channel.send_inputs("enable")

print("***** Get Prompt:")
prompt = conn.get_prompt()
print(prompt)

print("***** Show run | i hostname:")
result = conn.send_command("show run | i hostname")
print(result, result.result)

print("***** Disable Paging:")
result = conn.send_command("term length 0")
print(result, result.result)

print("***** Show run:")
result = conn.send_command("show run")
print(result, result.result)

if eos_device["keepalive"]:
    print("***** Waiting for keepalive....")
    time.sleep(5)

conn.close()



from datetime import datetime

from scrapli.response import Response


def test_response_init():
    response = Response("localhost", "ls -al")
    response_start_time = str(datetime.now())[:-7]
    assert response.host == "localhost"
    assert response.channel_input == "ls -al"
    assert str(response.start_time)[:-7] == response_start_time
    assert response.failed is True
    assert bool(response) is True
    assert repr(response) == "Scrape <Success: False>"
    assert str(response) == "Scrape <Success: False>"


def test_response_record_result():
    response = Response("localhost", "ls -al")
    response_end_time = str(datetime.now())[:-7]
    response_str = """
ls -al
total 264
drwxr-xr-x  34 carl  staff   1088 Jan 27 19:07 ./
drwxr-xr-x  21 carl  staff    672 Jan 25 15:56 ../
-rw-r--r--   1 carl  staff  53248 Jan 27 19:07 .coverage
drwxr-xr-x  12 carl  staff    384 Jan 27 19:13 .git/"""
    response.record_response(response_str)
    assert str(response.finish_time)[:-7] == response_end_time
    assert response.result == response_str


def test_response_parse_textfsm():
    response = Response("localhost", "show ip arp", "cisco_ios")
    response_str = """Protocol  Address          Age (min)  Hardware Addr   Type   Interface
Internet  172.31.254.1            -   0000.0c07.acfe  ARPA   Vlan254
Internet  172.31.254.2            -   c800.84b2.e9c2  ARPA   Vlan254
"""
    response.record_response(response_str)
    assert response.textfsm_parse_output()[0] == [
        "Internet",
        "172.31.254.1",
        "-",
        "0000.0c07.acfe",
        "ARPA",
        "Vlan254",
    ]


def test_response_parse_textfsm_fail():
    response = Response("localhost", "show ip arp", "cisco_ios")
    response_str = ""
    response.record_response(response_str)
    assert response.textfsm_parse_output() == []



import logging
import time
from pathlib import Path

from device_info import iosxe_device
from scrapli.driver.core import IOSXEDriver

logging.basicConfig(
    filename=f"{Path(__file__).resolve().parents[0]}/iosxe_driver.log", level=logging.DEBUG
)
logger = logging.getLogger("scrapli")

args = {
    "host": iosxe_device["host"],
    "port": iosxe_device["port"],
    "auth_username": iosxe_device["auth_username"],
    "auth_password": iosxe_device["auth_password"],
    "auth_strict_key": False,
    "keepalive_interval": iosxe_device["keepalive_interval"],
    "transport": iosxe_device["transport"],
    "keepalive": iosxe_device["keepalive"],
}

conn = IOSXEDriver(**args)
conn.open()

print("***** Get Prompt:")
prompt = conn.get_prompt()
print(prompt)

print("***** Show run | i hostname:")
result = conn.send_command("show run | i hostname")
print(result, result.result)

print("***** Clear logging buffer:")
interact = ["clear logg", "Clear logging buffer [confirm]", "", prompt]
result = conn.send_interactive(interact)
print(result, result.result)

print("***** Show run:")
result = conn.send_command("show run")
print(result, result.result)

if iosxe_device["keepalive"]:
    print("***** Waiting for keepalive....")
    time.sleep(5)

conn.close()



from scrapli.transport import SystemSSHTransport


def test_creation():
    conn = SystemSSHTransport("localhost")
    assert conn.host == "localhost"
    assert conn.port == 22
    assert conn._isauthenticated is False



import pytest


def test__str(mocked_channel):
    conn = mocked_channel([])
    assert str(conn.channel) == "scrapli Channel Object"


def test__repr(mocked_channel):
    conn = mocked_channel([])
    assert (
        repr(conn.channel)
        == r"scrapli Channel {'comms_prompt_pattern': '^[a-z0-9.\\-@()/:]{1,32}[#>$]\\s*$', 'comms_return_char': '\n', "
        r"'comms_ansi': False, 'timeout_ops': 10}"
    )


@pytest.mark.parametrize(
    "attr_setup",
    [
        (
            "send_input",
            {"channel_input": ["stuff", "more stuff"]},
            TypeError,
            "`send_input` expects a single string, got <class 'list'>. to send a list of inputs use the `send_inputs` method instead",
        ),
        (
            "send_inputs",
            {"channel_inputs": "stuff"},
            TypeError,
            "`send_inputs` expects a list of strings, got <class 'str'>. to send a single input use the `send_input` method instead",
        ),
        (
            "send_inputs_interact",
            {"channel_inputs": "stuff"},
            TypeError,
            "`send_inputs_interact` expects a List, got <class 'str'>",
        ),
    ],
    ids=["send_input", "send_inputs", "send_inputs_interact"],
)
def test_exceptions_raised(attr_setup, mocked_channel):
    method_name = attr_setup[0]
    methd_args = attr_setup[1]
    method_exc = attr_setup[2]
    method_msg = attr_setup[3]
    with pytest.raises(method_exc) as exc:
        conn = mocked_channel([])
        method = getattr(conn.channel, method_name)
        method(**methd_args)
    assert str(exc.value) == method_msg


@pytest.mark.parametrize(
    "attr_setup",
    [
        ({"strip_prompt": True}, b"hostname 3560CX\n"),
        ({"strip_prompt": False}, b"hostname 3560CX\n3560CX#"),
    ],
    ids=["strip_prompt_true", "strip_prompt_false",],
)
def test__restructure_output(attr_setup, mocked_channel):
    args = attr_setup[0]
    expected = attr_setup[1]
    channel_output = b"hostname 3560CX\r\n3560CX#"
    conn = mocked_channel([])
    output = conn.channel._restructure_output(channel_output, **args)
    assert output == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        ({"comms_ansi": False}, b"3560CX#", b"3560CX#"),
        ({"comms_ansi": True}, b"\x1b[0;0m3560CX#\x1b[0;0m", b"3560CX#"),
    ],
    ids=["read_chunk", "read_chunk_strip_ansi",],
)
def test__read_chunk(attr_setup, mocked_channel):
    args = attr_setup[0]
    initial_bytes = attr_setup[1]
    expected = attr_setup[2]
    conn = mocked_channel([], initial_bytes=initial_bytes, **args)
    output = conn.channel._read_chunk()
    assert output == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        ({"comms_ansi": False}, b"3560CX#", b"3560CX#"),
        ({"comms_ansi": True}, b"\x1b[0;0m3560CX#\x1b[0;0m", b"3560CX#"),
    ],
    ids=["read_chunk", "read_chunk_strip_ansi",],
)
def test__read_until_input(attr_setup, mocked_channel):
    args = attr_setup[0]
    initial_bytes = attr_setup[1]
    expected = attr_setup[2]
    conn = mocked_channel([], initial_bytes=initial_bytes, **args)
    output = conn.channel._read_until_input(b"3560")
    assert output == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        ({}, b"hostname 3560CX\r\n3560CX#"),
        ({"comms_prompt_pattern": "^3560CX#$"}, b"hostname 3560CX\r\n3560CX#"),
        ({"comms_prompt_pattern": "3560CX#"}, b"hostname 3560CX\r\n3560CX#"),
    ],
    ids=[
        "default_comms_prompt_pattern",
        "regex_comms_prompt_pattern",
        "string_comms_prompt_pattern",
    ],
)
def test__read_until_prompt(attr_setup, mocked_channel):
    args = attr_setup[0]
    expected = attr_setup[1]
    channel_output = b"hostname 3560CX\r\n"
    conn = mocked_channel([], **args)
    output = conn.channel._read_until_prompt(channel_output)
    assert output == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        ({}, "3560CX#"),
        ({"comms_prompt_pattern": "^3560CX#$"}, "3560CX#"),
        ({"comms_prompt_pattern": "3560CX#"}, "3560CX#"),
    ],
    ids=[
        "default_comms_prompt_pattern",
        "regex_comms_prompt_pattern",
        "string_comms_prompt_pattern",
    ],
)
def test_get_prompt(attr_setup, mocked_channel):
    args = attr_setup[0]
    expected = attr_setup[1]
    conn = mocked_channel([("\n", "\n3560CX#")], **args)
    output = conn.channel.get_prompt()
    assert output == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        (
            {},
            "show ip access-lists",
            """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#""",
            """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any""",
        ),
        (
            {"strip_prompt": False},
            "show ip access-lists",
            """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#""",
            """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#""",
        ),
    ],
    ids=["strip_prompt_true", "strip_prompt_false",],
)
def test_send_input(attr_setup, mocked_channel):
    args = attr_setup[0]
    channel_input = attr_setup[1]
    channel_output = attr_setup[2]
    expected = attr_setup[3]
    conn = mocked_channel([(channel_input, channel_output)])
    output = conn.channel.send_input(channel_input, **args)
    assert output.result == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        (
            {},
            ["show ip access-lists", "show ip prefix-list"],
            [
                """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#""",
                """ip prefix-list pl_default: 1 entries
   seq 5 permit 0.0.0.0/0 le 32
3560CX#""",
            ],
            [
                """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any""",
                """ip prefix-list pl_default: 1 entries
   seq 5 permit 0.0.0.0/0 le 32""",
            ],
        ),
        (
            {"strip_prompt": False},
            ["show ip access-lists", "show ip prefix-list"],
            [
                """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#""",
                """ip prefix-list pl_default: 1 entries
   seq 5 permit 0.0.0.0/0 le 32
3560CX#""",
            ],
            [
                """Extended IP access list ext_acl_fw
    10 deny ip 0.0.0.0 0.255.255.255 any
    20 deny ip 10.0.0.0 0.255.255.255 any
    30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
    40 deny ip 127.0.0.0 0.255.255.255 any
    50 deny ip 169.254.0.0 0.0.255.255 any
    60 deny ip 172.16.0.0 0.15.255.255 any
    70 deny ip 192.0.0.0 0.0.0.255 any
    80 deny ip 192.0.2.0 0.0.0.255 any
    90 deny ip 192.168.0.0 0.0.255.255 any
    100 deny ip 198.18.0.0 0.1.255.255 any
    110 deny ip 198.51.100.0 0.0.0.255 any
    120 deny ip 203.0.113.0 0.0.0.255 any
    130 deny ip 224.0.0.0 15.255.255.255 any
    140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#""",
                """ip prefix-list pl_default: 1 entries
   seq 5 permit 0.0.0.0/0 le 32
3560CX#""",
            ],
        ),
    ],
    ids=["strip_prompt_true", "strip_prompt_false",],
)
def test_send_inputs(attr_setup, mocked_channel):
    args = attr_setup[0]
    channel_inputs = attr_setup[1]
    channel_outputs = attr_setup[2]
    expected_results = attr_setup[3]

    test_operations = []
    for channel_input, channel_output in zip(channel_inputs, channel_outputs):
        test_operations.append((channel_input, channel_output))

    conn = mocked_channel(test_operations)
    outputs = conn.channel.send_inputs(channel_inputs, **args)

    for expected, output in zip(expected_results, outputs):
        assert output.result == expected


@pytest.mark.parametrize(
    "attr_setup",
    [
        (
            {},
            ["clear logg", "Clear logging buffer [confirm]", "", "3560CX#"],
            "Clear logging buffer [confirm]\n\n3560CX#",
        ),
        (
            {"hidden_response": True},
            ["clear logg", "Clear logging buffer [confirm]", "", "3560CX#"],
            "Clear logging buffer [confirm]\n\n3560CX#",
        ),
    ],
    ids=["basic", "hidden_response"],
)
def test_send_inputs_interact(attr_setup, mocked_channel):
    args = attr_setup[0]
    interact = attr_setup[1]
    expected = attr_setup[2]
    test_operations = [[interact[0], interact[1]], [interact[2], interact[3]]]
    if "hidden_response" in args:
        test_operations[1][0] = ""
    conn = mocked_channel(test_operations)
    output = conn.channel.send_inputs_interact(interact, **args)
    assert output.result == expected



from io import BytesIO
from typing import Any, Callable, Dict, Optional, Tuple

import pytest

from scrapli.channel import Channel
from scrapli.driver import NetworkDriver, Scrape
from scrapli.driver.core.cisco_iosxe.driver import PRIVS
from scrapli.transport.transport import Transport


class MockScrape(Scrape):
    def __init__(self, channel_ops, initial_bytes, *args, **kwargs):
        self.channel_ops = channel_ops
        self.initial_bytes = initial_bytes
        super().__init__(*args, **kwargs)

    def _transport_factory(self, transport: str) -> Tuple[Callable[..., Any], Dict[str, Any]]:
        """
        Private factory method to produce mock transport class

        Args:
            transport: string name of transport class to use

        Returns:
            Transport: initialized Transport class

        Raises:
            N/A  # noqa

        """
        transport_class = MockTransport
        required_transport_args = (
            "comms_return_char",
            "channel_ops",
            "initial_bytes",
        )
        transport_args = {}
        for arg in required_transport_args:
            transport_args[arg] = self._initialization_args.get(arg)
        transport_args["channel_ops"] = self.channel_ops
        transport_args["initial_bytes"] = self.initial_bytes
        return transport_class, transport_args


class MockTransport(Transport):
    def __init__(self, comms_return_char, initial_bytes, channel_ops):
        super().__init__()
        self.comms_return_char = comms_return_char
        self.channel_ops = channel_ops
        self.fd = BytesIO(initial_bytes=initial_bytes)
        self.input_counter = 0
        self.return_counter = 0

    def open(self) -> None:
        return

    def close(self) -> None:
        return

    def isalive(self) -> bool:
        return True

    def read(self) -> bytes:
        return self.fd.read(65535)

    def write(self, channel_input: str) -> None:
        if channel_input == self.comms_return_char:
            cur_input, cur_output = self.channel_ops[self.input_counter]
            self.input_counter += 1
            self.return_counter += 1

            self.fd.write(cur_output.encode())

            if self.return_counter == 1:
                self.return_counter = 0
                self.fd.seek(-len(cur_output) - 1, 1)
            else:
                self.fd.seek(len(cur_input) - 1)
        seek_offset = len(channel_input)
        self.fd.write(channel_input.encode())
        self.fd.seek(-seek_offset, 1)

    def set_timeout(self, timeout: Optional[int] = None) -> None:
        return

    def _keepalive_standard(self) -> None:
        return


class MockNetworkDriver(MockScrape, NetworkDriver):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.privs = PRIVS
        self.auth_secondary = "password"

    def open(self):
        # Overriding "normal" network driver open method as we don't need to worry about disable
        # paging or pre login handles; ignoring this makes the mocked file read/write pieces simpler
        self.transport = self.transport_class(**self.transport_args)
        self.transport.open()
        self.channel = Channel(self.transport, **self.channel_args)


@pytest.fixture(scope="module")
def mocked_channel():
    def _create_mocked_channel(test_operations, initial_bytes=b"3560CX#", **kwargs):
        conn = MockScrape(
            host="localhost", channel_ops=test_operations, initial_bytes=initial_bytes, **kwargs
        )
        conn.open()
        return conn

    return _create_mocked_channel


@pytest.fixture(scope="module")
def mocked_network_driver():
    def _create_mocked_network_driver(test_operations, initial_bytes=b"3560CX#", **kwargs):
        conn = MockNetworkDriver(
            host="localhost", channel_ops=test_operations, initial_bytes=initial_bytes, **kwargs
        )
        conn.open()
        return conn

    return _create_mocked_network_driver



"""scrapli.driver.core"""
from scrapli.driver.core.arista_eos.driver import EOSDriver
from scrapli.driver.core.cisco_iosxe.driver import IOSXEDriver
from scrapli.driver.core.cisco_iosxr.driver import IOSXRDriver
from scrapli.driver.core.cisco_nxos.driver import NXOSDriver
from scrapli.driver.core.juniper_junos.driver import JunosDriver

__all__ = (
    "EOSDriver",
    "IOSXEDriver",
    "IOSXRDriver",
    "NXOSDriver",
    "JunosDriver",
)



"""scrapli.driver.core.cisco_nxos.driver"""
from typing import Any, Callable, Dict, Optional

from scrapli.driver import NetworkDriver
from scrapli.driver.network_driver import PrivilegeLevel


def nxos_on_open(conn: NetworkDriver) -> None:
    """
    NXOSDriver default on_open callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    conn.channel.send_input("terminal length 0")


def nxos_on_close(conn: NetworkDriver) -> None:
    """
    NXOSDriver default on_close callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    # write exit directly to the transport as channel would fail to find the prompt after sending
    # the exit command!
    conn.acquire_priv(conn.default_desired_priv)
    conn.transport.write("exit")
    conn.transport.write(conn.channel.comms_prompt_pattern)


NXOS_ARG_MAPPER = {
    "comms_prompt_regex": r"^[a-z0-9.\-@()/:]{1,32}[#>$]\s*$",
    "comms_return_char": "\n",
    "on_open": nxos_on_open,
    "on_close": nxos_on_close,
}

PRIVS = {
    "exec": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@()/:]{1,32}>\s?$",
            "exec",
            "",
            "",
            "privilege_exec",
            "enable",
            True,
            "Password:",
            True,
            0,
        )
    ),
    "privilege_exec": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@/:]{1,32}#\s?$",
            "privilege_exec",
            "exec",
            "disable",
            "configuration",
            "configure terminal",
            False,
            "",
            True,
            1,
        )
    ),
    "configuration": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@/:]{1,32}\(config[a-z0-9.\-@/:]{0,16}\)#\s?$",
            "configuration",
            "privilege_exec",
            "end",
            "",
            "",
            False,
            "",
            True,
            2,
        )
    ),
    "special_configuration": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@/:]{1,32}\(config[a-z0-9.\-@/:]{1,16}\)#\s?$",
            "special_configuration",
            "privilege_exec",
            "end",
            "",
            "",
            False,
            "",
            False,
            3,
        )
    ),
}


class NXOSDriver(NetworkDriver):
    def __init__(
        self,
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,32}[#>$]\s?$",
        on_open: Optional[Callable[..., Any]] = None,
        on_close: Optional[Callable[..., Any]] = None,
        auth_secondary: str = "",
        transport: str = "system",
        **kwargs: Dict[str, Any],
    ):
        """
        NXOSDriver Object

        Args:
            comms_prompt_pattern: raw string regex pattern -- preferably use `^` and `$` anchors!
                this is the single most important attribute here! if this does not match a prompt,
                scrapli will not work!
                IMPORTANT: regex search uses multi-line + case insensitive flags. multi-line allows
                for highly reliably matching for prompts however we do NOT strip trailing whitespace
                for each line, so be sure to add `\\s*` if your device needs that. This should be
                mostly sorted for you if using network drivers (i.e. `IOSXEDriver`). Lastly, the
                case insensitive is just a convenience factor so i can be lazy.
            on_open: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately after authentication is completed. Common use
                cases for this callable would be to disable paging or accept any kind of banner
                message that prompts a user upon connection
            on_close: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately prior to closing the underlying transport.
                Common use cases for this callable would be to save configurations prior to exiting,
                or to logout properly to free up vtys or similar.
            auth_secondary: password to use for secondary authentication (enable)
            transport: system|ssh2|paramiko|telnet -- type of transport to use for connection
                system uses system available ssh (/usr/bin/ssh)
                ssh2 uses ssh2-python
                paramiko uses... paramiko
                telnet uses telnetlib
                choice of driver depends on the features you need. in general system is easiest as
                it will just 'auto-magically' use your ssh config file ('~/.ssh/config' or
                '/etc/ssh/config_file'). ssh2 is very very fast as it is a thin wrapper around
                libssh2 however it is slightly feature limited. paramiko is slower than ssh2, but
                has more features built in (though scrapli does not expose/support them all).
                explicitly added here to allow for nicely checking if transport is telnet.
            **kwargs: keyword args to pass to inherited class(es)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A
        """
        if on_open is None:
            on_open = nxos_on_open
        if on_close is None:
            on_close = nxos_on_close

        _telnet = False
        if transport == "telnet":
            _telnet = True

        super().__init__(
            auth_secondary,
            comms_prompt_pattern=comms_prompt_pattern,
            on_open=on_open,
            on_close=on_close,
            transport=transport,
            **kwargs,
        )

        self.privs = PRIVS
        self.default_desired_priv = "privilege_exec"
        self.textfsm_platform = "cisco_nxos"

        if _telnet:
            self.transport.username_prompt = "login:"



from scrapli.transport import TelnetTransport


def test_creation():
    conn = TelnetTransport("localhost")
    assert conn.host == "localhost"
    assert conn.port == 23
    assert conn._isauthenticated is False



"""scrapli.helper"""
import os
import re
import warnings
from io import TextIOWrapper
from pathlib import Path
from typing import Any, Dict, List, Optional, Pattern, TextIO, Union

import pkg_resources  # pylint: disable=C0411


def get_prompt_pattern(prompt: str, class_prompt: str) -> Pattern[bytes]:
    """
    Return compiled prompt pattern

    Given a potential prompt and the Channel class' prompt, return compiled prompt pattern

    Args:
        prompt: bytes string to process
        class_prompt: Channel class' prompt pattern

    Returns:
        output: bytes string each line right stripped

    Raises:
        N/A

    """
    check_prompt = prompt or class_prompt
    if isinstance(check_prompt, str):
        bytes_check_prompt = check_prompt.encode()
    else:
        bytes_check_prompt = check_prompt
    if bytes_check_prompt.startswith(b"^") and bytes_check_prompt.endswith(b"$"):
        return re.compile(bytes_check_prompt, flags=re.M | re.I)
    return re.compile(re.escape(bytes_check_prompt))


def normalize_lines(output: bytes) -> bytes:
    r"""
    Normalize lines

    Split output lines to remove \r\n, rstrip each line and rejoin

    Args:
        output: bytes string to process

    Returns:
        bytes: bytes string each line right stripped

    Raises:
        N/A

    """
    return b"\n".join([line.rstrip() for line in output.splitlines()])


def strip_ansi(output: bytes) -> bytes:
    """
    Strip comms_ansi

    Strip comms_ansi characters from output

    Args:
        output: bytes from previous reads if needed

    Returns:
        bytes: bytes output read from channel with comms_ansi characters removed

    Raises:
        N/A

    """
    ansi_escape_pattern = re.compile(rb"\x1b(\[.*?[@-~]|\].*?(\x07|\x1b\\))")
    output = re.sub(ansi_escape_pattern, b"", output)
    return output


def _textfsm_get_template(platform: str, command: str) -> Optional[TextIO]:
    """
    Find correct TextFSM template based on platform and command executed

    Args:
        platform: ntc-templates device type; i.e. cisco_ios, arista_eos, etc.
        command: string of command that was executed (to find appropriate template)

    Returns:
        None or TextIO of opened template

    Raises:
        N/A

    """
    try:
        from textfsm.clitable import CliTable  # pylint: disable=C0415
    except ModuleNotFoundError as exc:
        err = f"Module '{exc.name}' not installed!"
        msg = f"***** {err} {'*' * (80 - len(err))}"
        fix = (
            f"To resolve this issue, install '{exc.name}'. You can do this in one of the following"
            " ways:\n"
            "1: 'pip install -r requirements-textfsm.txt'\n"
            "2: 'pip install scrapli[textfsm]'"
        )
        warning = "\n" + msg + "\n" + fix + "\n" + msg
        warnings.warn(warning)
        return None
    template_dir = pkg_resources.resource_filename("ntc_templates", "templates")
    cli_table = CliTable("index", template_dir)
    template_index = cli_table.index.GetRowMatch({"Platform": platform, "Command": command})
    if not template_index:
        return None
    template_name = cli_table.index.index[template_index]["Template"]
    template = open(f"{template_dir}/{template_name}")
    return template


def textfsm_parse(
    template: Union[str, TextIOWrapper], output: str
) -> Union[List[Any], Dict[str, Any]]:
    """
    Parse output with TextFSM and ntc-templates, try to return structured output

    Args:
        template: TextIOWrapper or string path to template to use to parse data
        output: unstructured output from device to parse

    Returns:
        output: structured data

    Raises:
        N/A

    """
    import textfsm  # pylint: disable=C0415

    if not isinstance(template, TextIOWrapper):
        template_file = open(template)
    else:
        template_file = template
    re_table = textfsm.TextFSM(template_file)
    try:
        structured_output: Union[List[Any], Dict[str, Any]] = re_table.ParseText(output)
        return structured_output
    except textfsm.parser.TextFSMError:
        pass
    return []


def resolve_ssh_config(ssh_config_file: str) -> str:
    """
    Resolve ssh configuration file from provided string

    If provided string is empty (`""`) try to resolve system ssh config files located at
    `~/.ssh/config` or `/etc/ssh/ssh_config`.

    Args:
        ssh_config_file: string representation of ssh config file to try to use

    Returns:
        str: string to path fro ssh config file or an empty string

    Raises:
        N/A

    """
    if Path(ssh_config_file).is_file():
        return str(Path(ssh_config_file))
    if Path(os.path.expanduser("~/.ssh/config")).is_file():
        return str(Path(os.path.expanduser("~/.ssh/config")))
    if Path("/etc/ssh/ssh_config").is_file():
        return str(Path("/etc/ssh/ssh_config"))
    return ""


def resolve_ssh_known_hosts(ssh_known_hosts: str) -> str:
    """
    Resolve ssh known hosts file from provided string

    If provided string is empty (`""`) try to resolve system known hosts files located at
    `~/.ssh/known_hosts` or `/etc/ssh/ssh_known_hosts`.

    Args:
        ssh_known_hosts: string representation of ssh config file to try to use

    Returns:
        str: string to path fro ssh config file or an empty string

    Raises:
        N/A

    """
    if Path(ssh_known_hosts).is_file():
        return str(Path(ssh_known_hosts))
    if Path(os.path.expanduser("~/.ssh/known_hosts")).is_file():
        return str(Path(os.path.expanduser("~/.ssh/known_hosts")))
    if Path("/etc/ssh/ssh_known_hosts").is_file():
        return str(Path("/etc/ssh/ssh_known_hosts"))
    return ""



import os
import re
import sys
from io import TextIOWrapper
from pathlib import Path

import pkg_resources  # pylint: disable=C041
import pytest

import scrapli
from scrapli.helper import (
    _textfsm_get_template,
    get_prompt_pattern,
    resolve_ssh_config,
    strip_ansi,
    textfsm_parse,
)

UNIT_TEST_DIR = f"{Path(scrapli.__file__).parents[1]}/tests/unit/"

IOS_ARP = """Protocol  Address          Age (min)  Hardware Addr   Type   Interface
Internet  172.31.254.1            -   0000.0c07.acfe  ARPA   Vlan254
Internet  172.31.254.2            -   c800.84b2.e9c2  ARPA   Vlan254
"""


def test_get_prompt_pattern_class_pattern():
    class_pattern = "^averygoodpattern$"
    result = get_prompt_pattern("", class_pattern)
    assert result == re.compile(b"^averygoodpattern$", re.IGNORECASE | re.MULTILINE)


def test_get_prompt_pattern_class_pattern_no_line_start_end_markers():
    class_pattern = "averygoodpattern"
    result = get_prompt_pattern("", class_pattern)
    assert result == re.compile(b"averygoodpattern")


def test_get_prompt_pattern_arg_pattern():
    class_pattern = "averygoodpattern"
    result = get_prompt_pattern("^awesomepattern$", class_pattern)
    assert result == re.compile(b"^awesomepattern$", re.IGNORECASE | re.MULTILINE)


def test_get_prompt_pattern_arg_string():
    class_pattern = "averygoodpattern"
    result = get_prompt_pattern("awesomepattern", class_pattern)
    assert result == re.compile(b"awesomepattern")


def test__strip_ansi():
    output = b"[admin@CoolDevice.Sea1: \x1b[1m/\x1b[0;0m]$"
    output = strip_ansi(output)
    assert output == b"[admin@CoolDevice.Sea1: /]$"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test__textfsm_get_template_valid_template():
    template = _textfsm_get_template("cisco_nxos", "show ip arp")
    template_dir = pkg_resources.resource_filename("ntc_templates", "templates")
    assert isinstance(template, TextIOWrapper)
    assert template.name == f"{template_dir}/cisco_nxos_show_ip_arp.textfsm"


def test__textfsm_get_template_invalid_template():
    template = _textfsm_get_template("cisco_nxos", "show racecar")
    assert not template


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test_text_textfsm_parse_success():
    template = _textfsm_get_template("cisco_ios", "show ip arp")
    result = textfsm_parse(template, IOS_ARP)
    assert isinstance(result, list)
    assert result[0] == ["Internet", "172.31.254.1", "-", "0000.0c07.acfe", "ARPA", "Vlan254"]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test_text_textfsm_parse_success_string_path():
    template = _textfsm_get_template("cisco_ios", "show ip arp")
    result = textfsm_parse(template.name, IOS_ARP)
    assert isinstance(result, list)
    assert result[0] == ["Internet", "172.31.254.1", "-", "0000.0c07.acfe", "ARPA", "Vlan254"]


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting textfsm on windows")
def test_text_textfsm_parse_failure():
    template = _textfsm_get_template("cisco_ios", "show ip arp")
    result = textfsm_parse(template, "not really arp data")
    assert result == []


def test_resolve_ssh_config_file_explicit():
    ssh_conf = resolve_ssh_config(f"{UNIT_TEST_DIR}_ssh_config")
    assert ssh_conf == f"{UNIT_TEST_DIR}_ssh_config"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting ssh config on windows")
def test_resolve_ssh_config_file_user(fs):
    fs.add_real_file("/etc/ssh/ssh_config", target_path=f"{os.path.expanduser('~')}/.ssh/config")
    ssh_conf = resolve_ssh_config("")
    assert ssh_conf == f"{os.path.expanduser('~')}/.ssh/config"


@pytest.mark.skipif(sys.platform.startswith("win"), reason="not supporting ssh config on windows")
def test_resolve_ssh_config_file_system(fs):
    fs.add_real_file("/etc/ssh/ssh_config")
    ssh_conf = resolve_ssh_config("")
    assert ssh_conf == "/etc/ssh/ssh_config"



"""scrapli.driver.core.cisco_iosxr.driver"""
import time
from typing import Any, Callable, Dict, Optional

from scrapli.driver import NetworkDriver
from scrapli.driver.network_driver import PrivilegeLevel


def iosxr_on_open(conn: NetworkDriver) -> None:
    """
    IOSXRDriver default on_open callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    # sleep for session to establish; without this we never find base prompt for some reason?
    # maybe this is an artifact from previous iterations/tests and can be done away with...
    time.sleep(1)
    conn.acquire_priv(conn.default_desired_priv)
    conn.channel.send_input("terminal length 0")


def iosxr_on_close(conn: NetworkDriver) -> None:
    """
    IOSXRDriver default on_close callable

    Args:
        conn: NetworkDriver object

    Returns:
        N/A  # noqa: DAR202

    Raises:
        N/A
    """
    # write exit directly to the transport as channel would fail to find the prompt after sending
    # the exit command!
    conn.transport.write("exit")
    conn.transport.write(conn.channel.comms_prompt_pattern)


IOSXR_ARG_MAPPER = {
    "comms_prompt_regex": r"^[a-z0-9.\-@()/:]{1,32}[#>$]\s?$",
    "comms_return_char": "\n",
    "on_open": iosxr_on_open,
    "on_close": iosxr_on_close,
}

PRIVS = {
    "privilege_exec": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@/:]{1,32}#\s?$",
            "privilege_exec",
            "exec",
            "disable",
            "configuration",
            "configure terminal",
            False,
            "",
            True,
            1,
        )
    ),
    "configuration": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@/:]{1,32}\(config[a-z0-9.\-@/:]{0,16}\)#\s?$",
            "configuration",
            "privilege_exec",
            "end",
            "",
            "",
            False,
            "",
            True,
            2,
        )
    ),
    "special_configuration": (
        PrivilegeLevel(
            r"^[a-z0-9.\-@/:]{1,32}\(config[a-z0-9.\-@/:]{1,16}\)#\s?$",
            "special_configuration",
            "privilege_exec",
            "end",
            "",
            "",
            False,
            "",
            False,
            3,
        )
    ),
}


class IOSXRDriver(NetworkDriver):
    def __init__(
        self,
        comms_prompt_pattern: str = r"^[a-z0-9.\-@/:]{1,32}#\s?$",
        on_open: Optional[Callable[..., Any]] = None,
        on_close: Optional[Callable[..., Any]] = None,
        auth_secondary: str = "",
        **kwargs: Dict[str, Any],
    ):
        """
        IOSXRDriver Object

        Args:
            comms_prompt_pattern: raw string regex pattern -- preferably use `^` and `$` anchors!
                this is the single most important attribute here! if this does not match a prompt,
                scrapli will not work!
                IMPORTANT: regex search uses multi-line + case insensitive flags. multi-line allows
                for highly reliably matching for prompts however we do NOT strip trailing whitespace
                for each line, so be sure to add '\\s*' if your device needs that. This should be
                mostly sorted for you if using network drivers (i.e. `IOSXEDriver`). Lastly, the
                case insensitive is just a convenience factor so i can be lazy.
            on_open: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately after authentication is completed. Common use
                cases for this callable would be to disable paging or accept any kind of banner
                message that prompts a user upon connection
            on_close: callable that accepts the class instance as its only argument. this callable,
                if provided, is executed immediately prior to closing the underlying transport.
                Common use cases for this callable would be to save configurations prior to exiting,
                or to logout properly to free up vtys or similar.
            auth_secondary: password to use for secondary authentication (enable)
            **kwargs: keyword args to pass to inherited class(es)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A
        """
        if on_open is None:
            on_open = iosxr_on_open
        if on_close is None:
            on_close = iosxr_on_close

        super().__init__(
            auth_secondary,
            comms_prompt_pattern=comms_prompt_pattern,
            on_open=on_open,
            on_close=on_close,
            **kwargs,
        )

        self.privs = PRIVS
        self.default_desired_priv = "privilege_exec"
        self.textfsm_platform = "cisco_xr"



import os

import pytest
from napalm import get_network_driver

from .test_data.devices import DEVICES

NAPALM_DEVICE_TYPE_MAP = {
    "cisco_iosxe": "ios",
    "cisco_nxos": "nxos",
    "cisco_iosxr": "iosxr",
    "arista_eos": "eos",
    "juniper_junos": "junos",
}


@pytest.fixture(scope="session", autouse=True)
def prepare_device(request):
    if os.getenv("SCRAPLI_NO_SETUP", "").lower() == "true":
        return

    test_devices = []
    for test in request.node.items:
        # for linux testing we wont have a device_type to test against, and obviously dont need
        #  napalm to push configs!
        if "device_type" not in test.callspec.params:
            continue
        test_devices.append(test.callspec.params["device_type"])
    test_devices = set(test_devices)

    # push base config via napalm to ensure consistent testing experience
    for device in test_devices:
        base_config = DEVICES[device]["base_config"]

        napalm_device_type = NAPALM_DEVICE_TYPE_MAP.get(device)
        napalm_driver = get_network_driver(napalm_device_type)
        napalm_conn = napalm_driver(
            hostname=DEVICES[device]["host"],
            username=DEVICES[device]["auth_username"],
            password=DEVICES[device]["auth_password"],
        )
        napalm_conn.open()
        napalm_conn.load_replace_candidate(filename=base_config)
        napalm_conn.commit_config()
        napalm_conn.close()


# @pytest.fixture(scope="class", params=["cisco_iosxe", "cisco_nxos", "cisco_iosxr", "arista_eos", "juniper_junos"])
@pytest.fixture(
    scope="class",
    params=["cisco_iosxe", "cisco_nxos", "cisco_iosxr", "arista_eos", "juniper_junos"],
)
def device_type(request):
    yield request.param


# @pytest.fixture(scope="class", params=["system", "ssh2", "paramiko", "telnet"])
@pytest.fixture(scope="class", params=["system", "ssh2", "paramiko", "telnet"])
def transport(request):
    yield request.param


@pytest.fixture(scope="class")
def nix_conn(transport):
    if transport == "telnet":
        pytest.skip("skipping telnet for linux hosts")

    device = DEVICES["linux"].copy()
    driver = device.pop("driver")

    conn = driver(**device, transport=transport,)
    conn.open()
    return conn


@pytest.fixture(scope="class")
def conn(device_type, transport):
    if device_type == "arista_eos" and transport == "ssh2":
        pytest.skip(
            "SSH2 (on pypi) doesn't support keyboard interactive auth, skipping ssh2 for arista_eos testing"
        )

    device = DEVICES[device_type].copy()
    driver = device.pop("driver")
    device.pop("base_config")

    timeout_transport = 5
    timeout_ops = 5
    if device_type == "juniper_junos":
        # commits on vsrx take one whole eternity...
        timeout_transport = 30
        timeout_ops = 30

    port = 22
    if transport == "telnet":
        port = 23

    conn = driver(
        **device,
        port=port,
        transport=transport,
        timeout_socket=5,
        timeout_transport=timeout_transport,
        timeout_ops=timeout_ops,
    )
    conn.open()
    return conn



"""scrapli.transport"""
from scrapli.transport.cssh2 import SSH2_TRANSPORT_ARGS, SSH2Transport
from scrapli.transport.miko import MIKO_TRANSPORT_ARGS, MikoTransport
from scrapli.transport.systemssh import SYSTEM_SSH_TRANSPORT_ARGS, SystemSSHTransport
from scrapli.transport.telnet import TELNET_TRANSPORT_ARGS, TelnetTransport
from scrapli.transport.transport import Transport

__all__ = (
    "Transport",
    "MikoTransport",
    "MIKO_TRANSPORT_ARGS",
    "SSH2Transport",
    "SSH2_TRANSPORT_ARGS",
    "SystemSSHTransport",
    "SYSTEM_SSH_TRANSPORT_ARGS",
    "TELNET_TRANSPORT_ARGS",
    "TelnetTransport",
)



"""scrapli.transport.systemssh"""
import os
import pty
import re
from logging import getLogger
from select import select
from subprocess import PIPE, Popen
from typing import TYPE_CHECKING, Optional, Union

from scrapli.decorators import operation_timeout
from scrapli.exceptions import ScrapliAuthenticationFailed, ScrapliTimeout
from scrapli.helper import get_prompt_pattern, strip_ansi
from scrapli.ssh_config import SSHConfig
from scrapli.transport.ptyprocess import PtyProcess
from scrapli.transport.transport import Transport

if TYPE_CHECKING:
    PopenBytes = Popen[bytes]  # pylint: disable=E1136

LOG = getLogger("transport")

SYSTEM_SSH_TRANSPORT_ARGS = (
    "auth_username",
    "auth_private_key",
    "auth_password",
    "auth_strict_key",
    "ssh_config_file",
    "ssh_known_hosts_file",
    "comms_prompt_pattern",
    "comms_return_char",
    "comms_ansi",
    "timeout_ops",
)


class SystemSSHTransport(Transport):
    def __init__(
        self,
        host: str = "",
        port: int = 22,
        auth_username: str = "",
        auth_private_key: str = "",
        auth_password: str = "",
        auth_strict_key: bool = True,
        timeout_socket: int = 5,
        timeout_transport: int = 5,
        timeout_ops: int = 10,
        timeout_exit: bool = True,
        keepalive: bool = False,
        keepalive_interval: int = 30,
        keepalive_type: str = "",
        keepalive_pattern: str = "\005",
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,32}[#>$]$",
        comms_return_char: str = "\n",
        comms_ansi: bool = False,
        ssh_config_file: str = "",
        ssh_known_hosts_file: str = "",
    ) -> None:
        """
        SystemSSHTransport Object

        Inherit from Transport ABC
        SSH2Transport <- Transport (ABC)

        If using this driver, and passing a ssh_config_file (or setting this argument to `True`),
        all settings in the ssh config file will be superseded by any arguments passed here!

        SystemSSHTransport *always* prefers public key auth if given the option! If auth_private_key
        is set in the provided arguments OR if ssh_config_file is passed/True and there is a key for
        ANY match (i.e. `*` has a key in ssh config file!!), we will use that key! If public key
        auth fails and a username and password is set (manually or by ssh config file), password
        auth will be attempted.

        Note that comms_prompt_pattern, comms_return_char and comms_ansi are only passed here to
        handle "in channel" authentication required by SystemSSH -- these are assigned to private
        attributes in this class and ignored after authentication. If you wish to modify these
        values on a "live" scrapli connection, modify them in the Channel object, i.e.
        `conn.channel.comms_prompt_pattern`. Additionally timeout_ops is passed and assigned to
        _timeout_ops to use the same timeout_ops that is used in Channel to decorate the
        authentication methods here.

        Args:
            host: host ip/name to connect to
            port: port to connect to
            auth_username: username for authentication
            auth_private_key: path to private key for authentication
            auth_password: password for authentication
            auth_strict_key: True/False to enforce strict key checking (default is True)
            timeout_socket: timeout for ssh session to start -- this directly maps to ConnectTimeout
                ssh argument; see `man ssh_config`
            timeout_transport: timeout for transport in seconds. since system ssh is using popen/pty
                we can't really set a timeout directly, so this value governs the time timeout
                decorator for the transport read and write methods
            timeout_ops: timeout for telnet channel operations in seconds -- this is also the
                timeout for finding and responding to username and password prompts at initial
                login. This is assigned to a private attribute and is ignored after authentication
                is completed.
            timeout_exit: True/False close transport if timeout encountered. If False and keepalives
                are in use, keepalives will prevent program from exiting so you should be sure to
                catch Timeout exceptions and handle them appropriately
            keepalive: whether or not to try to keep session alive
            keepalive_interval: interval to use for session keepalives
            keepalive_type: network|standard -- 'network' sends actual characters over the
                transport channel. This is useful for network-y type devices that may not support
                'standard' keepalive mechanisms. 'standard' is not currently implemented for
                system ssh
            keepalive_pattern: pattern to send to keep network channel alive. Default is
                u'\005' which is equivalent to 'ctrl+e'. This pattern moves cursor to end of the
                line which should be an innocuous pattern. This will only be entered *if* a lock
                can be acquired. This is only applicable if using keepalives and if the keepalive
                type is 'network'
            comms_prompt_pattern: prompt pattern expected for device, same as the one provided to
                channel -- system ssh needs to know this to know how to decide if we are properly
                sending/receiving data -- i.e. we are not stuck at some password prompt or some
                other failure scenario. If using driver, this should be passed from driver (Scrape,
                or IOSXE, etc.) to this Transport class. This is assigned to a private attribute and
                is ignored after authentication is completed.
            comms_return_char: return character to use on the channel, same as the one provided to
                channel -- system ssh needs to know this to know what to send so that we can probe
                the channel to make sure we are authenticated and sending/receiving data. If using
                driver, this should be passed from driver (Scrape, or IOSXE, etc.) to this Transport
                class. This is assigned to a private attribute and is ignored after authentication
                is completed.
            comms_ansi: True/False strip comms_ansi characters from output; this value is assigned
                self._comms_ansi and is ignored after authentication. We only need it for transport
                on the off chance (maybe never?) that username/password prompts contain ansi
                characters, otherwise "comms_ansi" is really a channel attribute and is treated as
                such. This is assigned to a private attribute and is ignored after authentication
                is completed.
            ssh_config_file: string to path for ssh config file
            ssh_known_hosts_file: string to path for ssh known hosts file

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(
            host,
            port,
            timeout_socket,
            timeout_transport,
            timeout_exit,
            keepalive,
            keepalive_interval,
            keepalive_type,
            keepalive_pattern,
        )

        self.auth_username: str = auth_username
        self.auth_private_key: str = auth_private_key
        self.auth_password: str = auth_password
        self.auth_strict_key: bool = auth_strict_key

        self._timeout_ops: int = timeout_ops
        self._comms_prompt_pattern: str = comms_prompt_pattern
        self._comms_return_char: str = comms_return_char
        self._comms_ansi: bool = comms_ansi
        self._process_ssh_config(ssh_config_file)
        self.ssh_known_hosts_file: str = ssh_known_hosts_file

        self.session: Union[Popen[bytes], PtyProcess]  # pylint: disable=E1136
        self.lib_auth_exception = ScrapliAuthenticationFailed
        self._isauthenticated = False

        self.open_cmd = ["ssh", self.host]
        self._build_open_cmd()

        # create stdin/stdout fd in case we can use pipes for session
        self._stdin_fd = -1
        self._stdout_fd = -1

    def _process_ssh_config(self, ssh_config_file: str) -> None:
        """
        Method to parse ssh config file

        Ensure ssh_config_file is valid (if providing a string path to config file), or resolve
        config file if passed True. Search config file for any private key, if ANY matching key is
        found and user has not provided a private key, set `auth_private_key` to the value of the
        found key. This is because we prefer to use `open_pipes` over `open_pty`!

        Args:
            ssh_config_file: string path to ssh config file; passed down from `Scrape`, or the
                `NetworkDriver` or subclasses of it, in most cases.

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        ssh = SSHConfig(ssh_config_file)
        self.ssh_config_file = ssh.ssh_config_file
        host_config = ssh.lookup(self.host)
        if not self.auth_private_key and host_config.identity_file:
            self.auth_private_key = os.path.expanduser(host_config.identity_file.strip())

    def _build_open_cmd(self) -> None:
        """
        Method to craft command to open ssh session

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.open_cmd.extend(["-p", str(self.port)])
        self.open_cmd.extend(["-o", f"ConnectTimeout={self.timeout_socket}"])
        if self.auth_private_key:
            self.open_cmd.extend(["-i", self.auth_private_key])
        if self.auth_username:
            self.open_cmd.extend(["-l", self.auth_username])
        if self.auth_strict_key is False:
            self.open_cmd.extend(["-o", "StrictHostKeyChecking=no"])
            self.open_cmd.extend(["-o", "UserKnownHostsFile=/dev/null"])
        else:
            self.open_cmd.extend(["-o", "StrictHostKeyChecking=yes"])
            if self.ssh_known_hosts_file:
                self.open_cmd.extend(["-o", f"UserKnownHostsFile={self.ssh_known_hosts_file}"])
        if self.ssh_config_file:
            self.open_cmd.extend(["-F", self.ssh_config_file])
        else:
            self.open_cmd.extend(["-F", "/dev/null"])

    def open(self) -> None:
        """
        Parent method to open session, authenticate and acquire shell

        If possible it is preferable to use the `_open_pipes` method, but we can only do this IF we
        can authenticate with public key authorization (because we don't have to spawn a PTY; if no
        public key we have to spawn PTY to deal w/ authentication prompts). IF we get a private key
        provided, use pipes method, otherwise we will just deal with `_open_pty`. `_open_pty` is
        less preferable because we have to spawn a PTY and cannot as easily tell if SSH
        authentication is successful. With `_open_pipes` we can read stderr which contains the
        output from the verbose flag for SSH -- this contains a message that indicates success of
        SSH auth. In the case of `_open_pty` we have to read from the channel directly like in the
        case of telnet... so it works, but its just a bit less desirable.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliAuthenticationFailed: if all authentication means fail

        """
        self.session_lock.acquire()

        # If authenticating with private key prefer to use open pipes
        # _open_pipes uses subprocess Popen which is preferable to opening a pty
        # if _open_pipes fails and no password available, raise failure, otherwise try password auth
        if self.auth_private_key:
            open_pipes_result = self._open_pipes()
            if open_pipes_result:
                return
            if not self.auth_password or not self.auth_username:
                msg = (
                    f"Public key authentication to host {self.host} failed. Missing username or"
                    " password unable to attempt password authentication."
                )
                LOG.critical(msg)
                raise ScrapliAuthenticationFailed(msg)

        # If public key auth fails or is not configured, open a pty session
        if not self._open_pty():
            msg = f"Authentication to host {self.host} failed"
            LOG.critical(msg)
            raise ScrapliAuthenticationFailed(msg)

        if self.keepalive:
            self._session_keepalive()

    def _open_pipes(self) -> bool:
        """
        Private method to open session with subprocess.Popen

        Args:
            N/A

        Returns:
            bool: True/False session was opened and authenticated

        Raises:
            N/A

        """
        # copy the open_cmd as we don't want to update the objects open_cmd until we know we can
        # authenticate. add verbose output and disable batch mode (disables passphrase/password
        # queries). If auth is successful update the object open_cmd to represent what was used
        open_cmd = self.open_cmd.copy()
        open_cmd.append("-v")
        open_cmd.extend(["-o", "BatchMode=yes"])

        stdout_master_pty, stdout_slave_pty = pty.openpty()
        stdin_master_pty, stdin_slave_pty = pty.openpty()

        self.session = Popen(
            open_cmd,
            bufsize=0,
            shell=False,
            stdin=stdin_slave_pty,
            stdout=stdout_slave_pty,
            stderr=PIPE,
        )
        # close the slave fds, don't need them anymore
        os.close(stdin_slave_pty)
        os.close(stdout_slave_pty)
        LOG.debug(f"Session to host {self.host} spawned")

        try:
            self._pipes_isauthenticated(self.session)
        except TimeoutError:
            # If auth fails, kill the popen session, also need to manually close the stderr pipe
            # for some reason... unclear why, but w/out this it will hang open
            if self.session.stderr is not None:
                stderr_fd = self.session.stderr.fileno()
                os.close(stderr_fd)
            self.session.kill()
            return False

        LOG.debug(f"Authenticated to host {self.host} with public key")

        # set stdin/stdout to the new master pty fds
        self._stdin_fd = stdin_master_pty
        self._stdout_fd = stdout_master_pty

        self.open_cmd = open_cmd
        self.session_lock.release()
        return True

    @operation_timeout("_timeout_ops", "Timed out determining if session is authenticated")
    def _pipes_isauthenticated(self, pipes_session: "PopenBytes") -> bool:
        """
        Private method to check initial authentication when using subprocess.Popen

        Since we always run ssh with `-v` we can simply check the stderr (where verbose output goes)
        to see if `Authenticated to [our host]` is in the output.

        Args:
            pipes_session: Popen pipes session object

        Returns:
            bool: True/False session was authenticated

        Raises:
            ScrapliTimeout: if `Operation timed out` in stderr output

        """
        if pipes_session.stderr is None:
            raise ScrapliTimeout(f"Could not read stderr while connecting to host {self.host}")

        output = b""
        while True:
            output += pipes_session.stderr.read(65535)
            if f"Authenticated to {self.host}".encode() in output:
                self._isauthenticated = True
                return True
            if "Operation timed out".encode() in output:
                raise ScrapliTimeout(f"Timed opening connection to host {self.host}")

    def _open_pty(self) -> bool:
        """
        Private method to open session with PtyProcess

        Args:
            N/A

        Returns:
            bool: True/False session was opened and authenticated

        Raises:
            N/A

        """
        self.session = PtyProcess.spawn(self.open_cmd)
        LOG.debug(f"Session to host {self.host} spawned")
        self.session_lock.release()
        self._pty_authenticate(self.session)
        if not self._pty_isauthenticated(self.session):
            return False
        LOG.debug(f"Authenticated to host {self.host} with password")
        return True

    @operation_timeout("_timeout_ops", "Timed out looking for SSH login password prompt")
    def _pty_authenticate(self, pty_session: PtyProcess) -> None:
        """
        Private method to check initial authentication when using pty_session

        Args:
            pty_session: PtyProcess session object

        Returns:
            N/A  # noqa: DAR202

        Raises:
            ScrapliAuthenticationFailed: if we receive an EOFError -- this usually indicates that
                host key checking is enabled and failed.

        """
        self.session_lock.acquire()
        output = b""
        while True:
            try:
                output += pty_session.read()
            except EOFError:
                raise ScrapliAuthenticationFailed(f"Failed to open connection to host {self.host}")
            if self._comms_ansi:
                output = strip_ansi(output)
            if b"password" in output.lower():
                LOG.debug("Found password prompt, sending password")
                pty_session.write(self.auth_password.encode())
                pty_session.write(self._comms_return_char.encode())
                self.session_lock.release()
                break

    @operation_timeout("_timeout_ops", "Timed out determining if session is authenticated")
    def _pty_isauthenticated(self, pty_session: PtyProcess) -> bool:
        """
        Check if session is authenticated

        This is very naive -- it only knows if the sub process is alive and has not received an EOF.
        Beyond that we lock the session and send the return character and re-read the channel.

        Args:
            pty_session: PtyProcess session object

        Returns:
            bool: True if authenticated, else False

        Raises:
            N/A

        """
        LOG.debug("Attempting to determine if PTY authentication was successful")
        if pty_session.isalive() and not pty_session.eof():
            prompt_pattern = get_prompt_pattern("", self._comms_prompt_pattern)
            self.session_lock.acquire()
            pty_session.write(self._comms_return_char.encode())
            fd_ready, _, _ = select([pty_session.fd], [], [], 0)
            if pty_session.fd in fd_ready:
                output = b""
                while True:
                    new_output = pty_session.read()
                    output += new_output
                    # we do not need to deal w/ line replacement for the actual output, only for
                    # parsing if a prompt-like thing is at the end of the output
                    output = re.sub(b"\r", b"", output)
                    if self._comms_ansi:
                        output = strip_ansi(output)
                    channel_match = re.search(prompt_pattern, output)
                    if channel_match:
                        self.session_lock.release()
                        self._isauthenticated = True
                        return True
                    if b"password" in new_output.lower():
                        # if we see "password" we know auth failed (hopefully in all scenarios!)
                        return False
                    if new_output:
                        LOG.debug(
                            f"Cannot determine if authenticated, \n\tRead: {repr(new_output)}"
                        )
        self.session_lock.release()
        return False

    def close(self) -> None:
        """
        Close session and socket

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.session_lock.acquire()
        if isinstance(self.session, Popen):
            self.session.kill()
        elif isinstance(self.session, PtyProcess):
            self.session.kill(1)
        LOG.debug(f"Channel to host {self.host} closed")
        self.session_lock.release()

    def isalive(self) -> bool:
        """
        Check if session is alive and session is authenticated

        Args:
            N/A

        Returns:
            bool: True if session is alive and session authenticated, else False

        Raises:
            N/A

        """
        if isinstance(self.session, Popen):
            if self.session.poll() is None and self._isauthenticated:
                return True
        elif isinstance(self.session, PtyProcess):
            if self.session.isalive() and self._isauthenticated and not self.session.eof():
                return True
        return False

    @operation_timeout("timeout_transport", "Transport timeout during read operation.")
    def read(self) -> bytes:
        """
        Read data from the channel

        Args:
            N/A

        Returns:
            bytes: bytes output as read from channel

        Raises:
            N/A

        """
        read_bytes = 65535
        if isinstance(self.session, Popen):
            return os.read(self._stdout_fd, read_bytes)
        if isinstance(self.session, PtyProcess):
            return self.session.read(read_bytes)
        return b""

    @operation_timeout("timeout_transport", "Transport timeout during write operation.")
    def write(self, channel_input: str) -> None:
        """
        Write data to the channel

        Args:
            channel_input: string to send to channel

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if isinstance(self.session, Popen):
            os.write(self._stdin_fd, channel_input.encode())
        elif isinstance(self.session, PtyProcess):
            self.session.write(channel_input.encode())

    def set_timeout(self, timeout: Optional[int] = None) -> None:
        """
        Set session timeout

        Note that this modifies the objects `timeout_transport` value directly as this value is
        what controls the timeout decorator for read/write methods. This is slightly different
        behavior from ssh2/paramiko/telnet in that those transports modify the session value and
        leave the objects `timeout_transport` alone.

        Args:
            timeout: timeout in seconds

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        if isinstance(timeout, int):
            set_timeout = timeout
        else:
            set_timeout = self.timeout_transport
        self.timeout_transport = set_timeout

    def _keepalive_standard(self) -> None:
        """
        Send "out of band" (protocol level) keepalives to devices.

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            NotImplementedError: always, because this is not implemented for telnet

        """
        raise NotImplementedError("No 'standard' keepalive mechanism for telnet.")



#!/usr/bin/env python
"""scrapli - ssh|telnet screen scraping client library"""
import setuptools

__author__ = "Carl Montanari"

with open("README.md", "r") as f:
    README = f.read()

setuptools.setup(
    name="scrapli",
    version="2020.03.21",
    author=__author__,
    author_email="carl.r.montanari@gmail.com",
    description="Screen scraping (ssh|telnet) client focused on network devices",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/carlmontanari/scrapli",
    packages=setuptools.find_packages(),
    install_requires=[],
    extras_require={
        "textfsm": ["textfsm>=1.1.0", "ntc-templates>=1.1.0"],
        "paramiko": ["paramiko>=2.6.0"],
        "ssh2": ["ssh2-python>=0.18.0-1"],
    },
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS",
    ],
    python_requires=">=3.6",
)



import time
from threading import Lock

import pytest

from scrapli.decorators import operation_timeout

from .conftest import MockTransport


class SlowClass:
    def __init__(self):
        self.timeout_test = 0.5
        self.timeout_exit = True
        self.session_lock = Lock()
        self.session_lock.acquire()
        self.transport = MockTransport("\n", b"", [])

    @operation_timeout("timeout_test")
    def slow_function(self):
        time.sleep(1)

    @operation_timeout("timeout_test")
    def fast_function(self):
        return "fast"

    @operation_timeout("non_existent_class_attr")
    def confused_function(self):
        return "fast"

    def close(self):
        return


def test_operation_timeout_timeout():
    slow = SlowClass()
    with pytest.raises(TimeoutError):
        slow.slow_function()


def test_operation_timeout_success():
    slow = SlowClass()
    result = slow.fast_function()
    assert result == "fast"


def test_operation_timeout_no_class_attr():
    slow = SlowClass()
    result = slow.confused_function()
    assert result == "fast"



"""scrapli.driver.network_driver"""
import logging
import re
import warnings
from abc import ABC, abstractmethod
from collections import namedtuple
from typing import Any, Callable, Dict, List, Tuple, Union

from scrapli.driver.driver import Scrape
from scrapli.exceptions import CouldNotAcquirePrivLevel, UnknownPrivLevel
from scrapli.helper import get_prompt_pattern
from scrapli.response import Response
from scrapli.transport import (
    MIKO_TRANSPORT_ARGS,
    SSH2_TRANSPORT_ARGS,
    SYSTEM_SSH_TRANSPORT_ARGS,
    MikoTransport,
    SSH2Transport,
    SystemSSHTransport,
    Transport,
)

TRANSPORT_CLASS: Dict[str, Callable[..., Transport]] = {
    "system": SystemSSHTransport,
    "ssh2": SSH2Transport,
    "paramiko": MikoTransport,
}
TRANSPORT_ARGS: Dict[str, Tuple[str, ...]] = {
    "system": SYSTEM_SSH_TRANSPORT_ARGS,
    "ssh2": SSH2_TRANSPORT_ARGS,
    "paramiko": MIKO_TRANSPORT_ARGS,
}


PrivilegeLevel = namedtuple(
    "PrivilegeLevel",
    "pattern "
    "name "
    "deescalate_priv "
    "deescalate "
    "escalate_priv "
    "escalate "
    "escalate_auth "
    "escalate_prompt "
    "requestable "
    "level",
)

NoPrivLevel = PrivilegeLevel("", "", "", "", "", "", "", "", "", "")


PRIVS: Dict[str, PrivilegeLevel] = {}

LOG = logging.getLogger("scrapli_base")


class NetworkDriver(Scrape, ABC):
    @abstractmethod
    def __init__(
        self, auth_secondary: str = "", **kwargs: Any,
    ):
        """
        BaseNetworkDriver Object

        Args:
            auth_secondary: password to use for secondary authentication (enable)
            **kwargs: keyword args to pass to inherited class(es)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        super().__init__(**kwargs)

        self.textfsm_platform: str = ""
        self.auth_secondary = auth_secondary
        self.privs = PRIVS
        self.default_desired_priv: str = ""
        self._current_priv_level = NoPrivLevel

    def _determine_current_priv(self, current_prompt: str) -> PrivilegeLevel:
        """
        Determine current privilege level from prompt string

        Args:
            current_prompt: string of current prompt

        Returns:
            PrivilegeLevel: NamedTuple of current privilege level

        Raises:
            UnknownPrivLevel: if privilege level cannot be determined

        """
        for priv_level in self.privs.values():
            prompt_pattern = get_prompt_pattern("", priv_level.pattern)
            if re.search(prompt_pattern, current_prompt.encode()):
                return priv_level
        raise UnknownPrivLevel

    def _escalate(self) -> None:
        """
        Escalate to the next privilege level up

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            UnknownPrivLevel: if priv level cant be attained
            TypeError: if invalid next prompt value

        """
        current_priv = self._determine_current_priv(self.channel.get_prompt())
        if not current_priv.escalate:
            return

        next_priv = self.privs.get(current_priv.escalate_priv, None)
        if next_priv is None:
            raise UnknownPrivLevel(
                f"Could not get next priv level, current priv is {current_priv.name}"
            )
        next_prompt = next_priv.pattern
        if current_priv.escalate_auth:
            if not self.auth_secondary:
                err = (
                    "Privilege escalation generally requires an `auth_secondary` password, "
                    "but none is set!"
                )
                msg = f"***** {err} {'*' * (80 - len(err))}"
                fix = (
                    "scrapli will try to escalate privilege without entering a password but may "
                    "fail.\nSet an `auth_secondary` password if your device requires a password to "
                    "increase privilege, otherwise ignore this message."
                )
                warning = "\n" + msg + "\n" + fix + "\n" + msg
                warnings.warn(warning)
            else:
                escalate_cmd: str = current_priv.escalate
                escalate_prompt: str = current_priv.escalate_prompt
                escalate_auth = self.auth_secondary
                if not isinstance(next_prompt, str):
                    raise TypeError(
                        f"got {type(next_prompt)} for {current_priv.name} escalate priv, "
                        "expected str"
                    )
                self.channel.send_inputs_interact(
                    [escalate_cmd, escalate_prompt, escalate_auth, next_prompt],
                    hidden_response=True,
                )
                self.channel.comms_prompt_pattern = next_priv.pattern
                return
        self.channel.comms_prompt_pattern = next_priv.pattern
        self.channel.send_input(current_priv.escalate)

    def _deescalate(self) -> None:
        """
        Deescalate to the next privilege level down

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            UnknownPrivLevel: if no default priv level set to deescalate to

        """
        current_priv = self._determine_current_priv(self.channel.get_prompt())
        if current_priv.deescalate:
            next_priv = self.privs.get(current_priv.deescalate_priv, None)
            if not next_priv:
                raise UnknownPrivLevel(
                    "NetworkDriver has no default priv levels, set them or use a network driver"
                )
            self.channel.comms_prompt_pattern = next_priv.pattern
            self.channel.send_input(current_priv.deescalate)

    def acquire_priv(self, desired_priv: str) -> None:
        """
        Acquire desired priv level

        Args:
            desired_priv: string name of desired privilege level
                (see scrapli.driver.<driver_category.device_type>.driver for levels)

        Returns:
            N/A  # noqa: DAR202

        Raises:
            CouldNotAcquirePrivLevel: if requested priv level not attained

        """
        priv_attempt_counter = 0
        while True:
            current_priv = self._determine_current_priv(self.channel.get_prompt())
            if current_priv == self.privs[desired_priv]:
                self._current_priv_level = current_priv
                return
            if priv_attempt_counter > len(self.privs):
                raise CouldNotAcquirePrivLevel(
                    f"Could not get to '{desired_priv}' privilege level."
                )
            if current_priv.level > self.privs[desired_priv].level:
                self._deescalate()
            else:
                self._escalate()
            priv_attempt_counter += 1

    def send_command(self, command: str, strip_prompt: bool = True) -> Response:
        """
        Send a command

        Args:
            command: string to send to device in privilege exec mode
            strip_prompt: True/False strip prompt from returned output

        Returns:
            Response: Scrapli Response object

        Raises:
            TypeError: if command is anything but a string

        """
        if not isinstance(command, str):
            raise TypeError(
                f"`send_command` expects a single string, got {type(command)}. "
                "to send a list of commands use the `send_commands` method instead."
            )

        if self._current_priv_level.name != self.default_desired_priv:
            self.acquire_priv(self.default_desired_priv)
        response = self.channel.send_input(command, strip_prompt)

        # update the response objects with textfsm platform; we do this here because the underlying
        #  channel doesn't know or care about platforms
        response.textfsm_platform = self.textfsm_platform

        return response

    def send_commands(self, commands: List[str], strip_prompt: bool = True) -> List[Response]:
        """
        Send multiple commands

        Args:
            commands: list of strings to send to device in privilege exec mode
            strip_prompt: True/False strip prompt from returned output

        Returns:
            responses: list of Scrapli Response objects

        Raises:
            TypeError: if commands is anything but a list

        """
        if not isinstance(commands, list):
            raise TypeError(
                f"`send_commands` expects a list of strings, got {type(commands)}. "
                "to send a single command use the `send_command` method instead."
            )

        if self._current_priv_level.name != self.default_desired_priv:
            self.acquire_priv(self.default_desired_priv)
        responses = self.channel.send_inputs(commands, strip_prompt)

        # update the response objects with textfsm platform; we do this here because the underlying
        #  channel doesn't know or care about platforms
        for response in responses:
            response.textfsm_platform = self.textfsm_platform

        return responses

    def send_interactive(self, interact: List[str], hidden_response: bool = False) -> Response:
        """
        Send inputs in an interactive fashion; used to handle prompts

        accepts inputs and looks for expected prompt;
        sends the appropriate response, then waits for the "finale"
        returns the results of the interaction

        could be "chained" together to respond to more than a "single" staged prompt

        Args:
            interact: list of four string elements representing...
                channel_input - initial input to send
                expected_prompt - prompt to expect after initial input
                response - response to prompt
                final_prompt - final prompt to expect
            hidden_response: True/False response is hidden (i.e. password input)

        Returns:
            Response: scrapli Response object

        Raises:
            N/A

        """
        if self._current_priv_level.name != self.default_desired_priv:
            self.acquire_priv(self.default_desired_priv)
        response = self.channel.send_inputs_interact(interact, hidden_response)
        return response

    def send_configs(
        self, configs: Union[str, List[str]], strip_prompt: bool = True
    ) -> List[Response]:
        """
        Send configuration(s)

        Args:
            configs: string or list of strings to send to device in config mode
            strip_prompt: True/False strip prompt from returned output

        Returns:
            responses: List of Scrape Response objects

        Raises:
            N/A

        """
        if isinstance(configs, str):
            configs = [configs]

        self.acquire_priv("configuration")
        responses = self.channel.send_inputs(configs, strip_prompt)
        self.acquire_priv(self.default_desired_priv)
        return responses

    def get_prompt(self) -> str:
        """
        Convenience method to get device prompt from Channel

        Args:
            N/A

        Returns:
            str: prompt received from channel.get_prompt

        Raises:
            N/A

        """
        prompt: str = self.channel.get_prompt()
        return prompt



import types

import pytest

from scrapli.driver.network_driver import PrivilegeLevel
from scrapli.exceptions import CouldNotAcquirePrivLevel, UnknownPrivLevel

try:
    import ntc_templates
    import textfsm

    textfsm_avail = True
except ImportError:
    textfsm_avail = False


def test__determine_current_priv(mocked_network_driver):
    conn = mocked_network_driver([])
    current_priv = conn._determine_current_priv("execprompt>")
    assert current_priv.name == "exec"


def test__determine_current_priv_unknown(mocked_network_driver):
    conn = mocked_network_driver([])
    with pytest.raises(UnknownPrivLevel):
        conn._determine_current_priv("!!!!thisissoooowrongggg!!!!!!?!")


def test__escalate(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "configure terminal"
    channel_output_2 = """Enter configuration commands, one per line.  End with CNTL/Z.
3560CX(config)#"""
    channel_ops = [(channel_input_1, channel_output_1), (channel_input_2, channel_output_2)]

    conn = mocked_network_driver(channel_ops)
    conn._escalate()


def test__escalate_unknown_priv(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "configure terminal"
    channel_output_2 = """Enter configuration commands, one per line.  End with CNTL/Z.
    3560CX(config)#"""
    channel_ops = [(channel_input_1, channel_output_1), (channel_input_2, channel_output_2)]

    conn = mocked_network_driver(channel_ops)

    mock_privs = {
        "privilege_exec": (
            PrivilegeLevel(
                r"^[a-z0-9.\-@/:]{1,32}#$",
                "privilege_exec",
                "exec",
                "disable",
                "configuration",
                "configure terminal",
                False,
                False,
                True,
                1,
            )
        ),
    }
    conn.privs = mock_privs

    with pytest.raises(UnknownPrivLevel):
        conn._escalate()


def test__escalate_auth(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX>"
    channel_input_2 = "enable"
    channel_output_2 = "Password:"
    channel_input_3 = "password"
    channel_output_3 = "\n3560CX#"
    channel_ops = [
        (channel_input_1, channel_output_1),
        (channel_input_2, channel_output_2),
        (channel_input_3, channel_output_3),
    ]

    conn = mocked_network_driver(channel_ops)

    mock_privs = {
        "exec": (
            PrivilegeLevel(
                r"^[a-z0-9.\-@()/:]{1,32}>$",
                "exec",
                None,
                None,
                "privilege_exec",
                "enable",
                True,
                "Password:",
                True,
                0,
            )
        ),
        "privilege_exec": (
            PrivilegeLevel(
                r"^[a-z0-9.\-@/:]{1,32}#$",
                "privilege_exec",
                "exec",
                "disable",
                "configuration",
                "configure terminal",
                False,
                False,
                True,
                1,
            )
        ),
    }
    conn.privs = mock_privs

    conn._escalate()


def test__deescalate(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "end"
    channel_output_2 = "3560CX>"
    channel_ops = [(channel_input_1, channel_output_1), (channel_input_2, channel_output_2)]

    conn = mocked_network_driver(channel_ops)
    conn._deescalate()


def test_acquire_priv(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX>"
    channel_input_2 = "\n"
    channel_output_2 = "3560CX>"
    channel_input_3 = "enable"
    channel_output_3 = "Password: "
    channel_input_4 = "password"
    channel_output_4 = "\n3560CX#"
    channel_input_5 = "\n"
    channel_output_5 = "3560CX#"
    channel_input_6 = "\n"
    channel_output_6 = "3560CX#"
    channel_input_7 = "configure terminal"
    channel_output_7 = """Enter configuration commands, one per line.  End with CNTL/Z.
3560CX(config)#"""
    channel_input_8 = "\n"
    channel_output_8 = "3560CX(config)#"
    channel_ops = [
        (channel_input_1, channel_output_1),
        (channel_input_2, channel_output_2),
        (channel_input_3, channel_output_3),
        (channel_input_4, channel_output_4),
        (channel_input_5, channel_output_5),
        (channel_input_6, channel_output_6),
        (channel_input_7, channel_output_7),
        (channel_input_8, channel_output_8),
    ]

    conn = mocked_network_driver(channel_ops)
    conn.acquire_priv("configuration")


def test_acquire_priv_deescalate(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "3560CX(config)#"
    channel_input_2 = "\n"
    channel_output_2 = "3560CX(config)#"
    channel_input_3 = "end"
    channel_output_3 = "3560CX#"
    channel_input_4 = "\n"
    channel_output_4 = "3560CX#"
    channel_ops = [
        (channel_input_1, channel_output_1),
        (channel_input_2, channel_output_2),
        (channel_input_3, channel_output_3),
        (channel_input_4, channel_output_4),
    ]

    conn = mocked_network_driver(channel_ops)
    conn.acquire_priv("privilege_exec")


def test_acquire_priv_could_not_acquire_priv(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX>"
    channel_input_2 = "\n"
    channel_output_2 = "3560CX>"
    channel_input_3 = "enable"
    channel_output_3 = "Password: "
    channel_input_4 = "password"
    channel_output_4 = "\n3560CX#"

    channel_input_5 = "\n"
    channel_output_5 = "\n3560CX>"
    channel_input_6 = "\n"
    channel_output_6 = "3560CX>"
    channel_input_7 = "enable"
    channel_output_7 = "Password: "
    channel_input_8 = "password"
    channel_output_8 = "\n3560CX#"

    channel_input_9 = "\n"
    channel_output_9 = "\n3560CX>"
    channel_input_10 = "\n"
    channel_output_10 = "3560CX>"
    channel_input_11 = "enable"
    channel_output_11 = "Password: "
    channel_input_12 = "password"
    channel_output_12 = "\n3560CX#"

    channel_input_13 = "\n"
    channel_output_13 = "\n3560CX>"

    channel_ops = [
        (channel_input_1, channel_output_1),
        (channel_input_2, channel_output_2),
        (channel_input_3, channel_output_3),
        (channel_input_4, channel_output_4),
        (channel_input_5, channel_output_5),
        (channel_input_6, channel_output_6),
        (channel_input_7, channel_output_7),
        (channel_input_8, channel_output_8),
        (channel_input_9, channel_output_9),
        (channel_input_10, channel_output_10),
        (channel_input_11, channel_output_11),
        (channel_input_12, channel_output_12),
        (channel_input_13, channel_output_13),
    ]

    conn = mocked_network_driver(channel_ops)
    mock_privs = {
        "exec": (
            PrivilegeLevel(
                r"^[a-z0-9.\-@()/:]{1,32}>$",
                "exec",
                None,
                None,
                "privilege_exec",
                "enable",
                True,
                "Password:",
                True,
                0,
            )
        ),
        "privilege_exec": (
            PrivilegeLevel(
                r"^[a-z0-9.\-@/:]{1,32}#$",
                "privilege_exec",
                "exec",
                "disable",
                "configuration",
                "configure terminal",
                False,
                False,
                True,
                1,
            )
        ),
    }
    conn.privs = mock_privs

    def _mock_escalate(self):
        self.__class__._escalate(self)
        self.channel.comms_prompt_pattern = mock_privs["exec"].pattern

    def _mock_get_prompt():
        return "3560CX>"

    conn._escalate = types.MethodType(_mock_escalate, conn)
    conn.get_prompt = _mock_get_prompt

    with pytest.raises(CouldNotAcquirePrivLevel) as exc:
        conn.acquire_priv("privilege_exec")
    assert str(exc.value) == "Could not get to 'privilege_exec' privilege level."


def test_send_commands(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "show ip access-lists"
    channel_output_2 = """Extended IP access list ext_acl_fw
        10 deny ip 0.0.0.0 0.255.255.255 any
        20 deny ip 10.0.0.0 0.255.255.255 any
        30 deny ip 100.64.0.0 0.63.255.255 any (2 matches)
        40 deny ip 127.0.0.0 0.255.255.255 any
        50 deny ip 169.254.0.0 0.0.255.255 any
        60 deny ip 172.16.0.0 0.15.255.255 any
        70 deny ip 192.0.0.0 0.0.0.255 any
        80 deny ip 192.0.2.0 0.0.0.255 any
        90 deny ip 192.168.0.0 0.0.255.255 any
        100 deny ip 198.18.0.0 0.1.255.255 any
        110 deny ip 198.51.100.0 0.0.0.255 any
        120 deny ip 203.0.113.0 0.0.0.255 any
        130 deny ip 224.0.0.0 15.255.255.255 any
        140 deny ip 240.0.0.0 15.255.255.255 any
3560CX#"""
    test_operations = [(channel_input_1, channel_output_1), (channel_input_2, channel_output_2)]
    conn = mocked_network_driver(test_operations)
    conn.default_desired_priv = "privilege_exec"
    output = conn.send_command(channel_input_2, strip_prompt=False)
    assert output.result == channel_output_2


def test_send_inputs_interact(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "3560CX#"
    channel_input_2 = "clear logg"
    channel_output_2 = "Clear logging buffer [confirm]"
    channel_input_3 = "\n"
    channel_output_3 = "3560CX#"
    interact = [channel_input_2, channel_output_2, "", channel_output_3]
    test_operations = [
        (channel_input_1, channel_output_1),
        (channel_input_2, channel_output_2),
        (channel_input_3, channel_output_3),
    ]
    conn = mocked_network_driver(test_operations)
    conn.default_desired_priv = "privilege_exec"
    output = conn.send_interactive(interact, hidden_response=False)
    assert output.result == "Clear logging buffer [confirm]\n\n3560CX#"


def test_send_configs(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "\n"
    channel_output_2 = "\n3560CX#"
    channel_input_3 = "configure terminal"
    channel_output_3 = """Enter configuration commands, one per line.  End with CNTL/Z.
3560CX(config)#"""
    channel_input_4 = "\n"
    channel_output_4 = "3560CX(config)#"
    channel_input_5 = "hostname XC0653"
    channel_output_5 = "XC0653(config)#"
    channel_input_6 = "\n"
    channel_output_6 = "XC0653(config)#"
    channel_input_7 = "\n"
    channel_output_7 = "XC0653(config)#"
    channel_input_8 = "end"
    channel_output_8 = "3560CX#"
    channel_input_9 = "\n"
    channel_output_9 = "\n3560CX#"
    test_operations = [
        (channel_input_1, channel_output_1),
        (channel_input_2, channel_output_2),
        (channel_input_3, channel_output_3),
        (channel_input_4, channel_output_4),
        (channel_input_5, channel_output_5),
        (channel_input_6, channel_output_6),
        (channel_input_7, channel_output_7),
        (channel_input_8, channel_output_8),
        (channel_input_9, channel_output_9),
    ]
    conn = mocked_network_driver(test_operations)
    conn.default_desired_priv = "privilege_exec"
    output = conn.send_configs(channel_input_5, strip_prompt=False)
    assert output[0].result == channel_output_5


@pytest.mark.skipif(textfsm_avail is False, reason="textfsm and/or ntc_templates are not installed")
def test_send_inputs_textfsm_success(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "show version"
    channel_output_2 = """Cisco IOS Software, C3560CX Software (C3560CX-UNIVERSALK9-M), Version 15.2(4)E7, RELEASE SOFTWARE (fc2)
Technical Support: http://www.cisco.com/techsupport
Copyright (c) 1986-2018 by Cisco Systems, Inc.
Compiled Tue 18-Sep-18 13:20 by prod_rel_team

ROM: Bootstrap program is C3560CX boot loader
BOOTLDR: C3560CX Boot Loader (C3560CX-HBOOT-M) Version 15.2(4r)E5, RELEASE SOFTWARE (fc4)

3560CX uptime is 2 weeks, 3 days, 18 hours, 3 minutes
System returned to ROM by power-on
System restarted at 15:59:35 PST Wed Jan 15 2020
System image file is "flash:c3560cx-universalk9-mz.152-4.E7.bin"
Last reload reason: power-on



This product contains cryptographic features and is subject to United
States and local country laws governing import, export, transfer and
use. Delivery of Cisco cryptographic products does not imply
third-party authority to import, export, distribute or use encryption.
Importers, exporters, distributors and users are responsible for
compliance with U.S. and local country laws. By using this product you
agree to comply with applicable laws and regulations. If you are unable
to comply with U.S. and local laws, return this product immediately.

A summary of U.S. laws governing Cisco cryptographic products may be found at:
http://www.cisco.com/wwl/export/crypto/tool/stqrg.html

If you require further assistance please contact us by sending email to
export@cisco.com.

License Level: ipservices
License Type: Permanent Right-To-Use
Next reload license Level: ipservices

cisco WS-C3560CX-8PC-S (APM86XXX) processor (revision A0) with 524288K bytes of memory.
Processor board ID FOCXXXXXXXX
Last reset from power-on
5 Virtual Ethernet interfaces
12 Gigabit Ethernet interfaces
The password-recovery mechanism is enabled.

512K bytes of flash-simulated non-volatile configuration memory.
Base ethernet MAC Address       : FF:FF:FF:FF:FF:FF
Motherboard assembly number     : XX-XXXXX-XX
Power supply part number        : XXX-XXXX-XX
Motherboard serial number       : FOCXXXXXXXX
Power supply serial number      : FOCXXXXXXXX
Model revision number           : A0
Motherboard revision number     : A0
Model number                    : WS-C3560CX-8PC-S
System serial number            : FOCXXXXXXXX
Top Assembly Part Number        : XX-XXXX-XX
Top Assembly Revision Number    : A0
Version ID                      : V01
CLEI Code Number                : XXXXXXXXXX
Hardware Board Revision Number  : 0x02


Switch Ports Model                     SW Version            SW Image
------ ----- -----                     ----------            ----------
*    1 12    WS-C3560CX-8PC-S          15.2(4)E7             C3560CX-UNIVERSALK9-M


Configuration register is 0xF

3560CX#"""
    test_operations = [(channel_input_1, channel_output_1), (channel_input_2, channel_output_2)]
    conn = mocked_network_driver(test_operations)
    conn.default_desired_priv = "privilege_exec"
    conn.textfsm_platform = "cisco_ios"
    results = conn.send_command(channel_input_2)
    parsed_results = results.textfsm_parse_output()
    assert isinstance(parsed_results, list)
    assert parsed_results[0][0] == "15.2(4)E7"


@pytest.mark.skipif(textfsm_avail is False, reason="textfsm and/or ntc_templates are not installed")
def test_send_inputs_textfsm_fail(mocked_network_driver):
    channel_input_1 = "\n"
    channel_output_1 = "\n3560CX#"
    channel_input_2 = "show version"
    channel_output_2 = """Cisco IOS Software, C3560CX Software (C3560CX-UNIVERSALK9-M), Version 15.2(4)E7, RELEASE SOFTWARE (fc2)
Technical Support: http://www.cisco.com/techsupport
Copyright (c) 1986-2018 by Cisco Systems, Inc.
Compiled Tue 18-Sep-18 13:20 by prod_rel_team

ROM: Bootstrap program is C3560CX boot loader
BOOTLDR: C3560CX Boot Loader (C3560CX-HBOOT-M) Version 15.2(4r)E5, RELEASE SOFTWARE (fc4)

3560CX uptime is 2 weeks, 3 days, 18 hours, 3 minutes
System returned to ROM by power-on
System restarted at 15:59:35 PST Wed Jan 15 2020
System image file is "flash:c3560cx-universalk9-mz.152-4.E7.bin"
Last reload reason: power-on



This product contains cryptographic features and is subject to United
States and local country laws governing import, export, transfer and
use. Delivery of Cisco cryptographic products does not imply
third-party authority to import, export, distribute or use encryption.
Importers, exporters, distributors and users are responsible for
compliance with U.S. and local country laws. By using this product you
agree to comply with applicable laws and regulations. If you are unable
to comply with U.S. and local laws, return this product immediately.

A summary of U.S. laws governing Cisco cryptographic products may be found at:
http://www.cisco.com/wwl/export/crypto/tool/stqrg.html

If you require further assistance please contact us by sending email to
export@cisco.com.

License Level: ipservices
License Type: Permanent Right-To-Use
Next reload license Level: ipservices

cisco WS-C3560CX-8PC-S (APM86XXX) processor (revision A0) with 524288K bytes of memory.
Processor board ID FOCXXXXXXXX
Last reset from power-on
5 Virtual Ethernet interfaces
12 Gigabit Ethernet interfaces
The password-recovery mechanism is enabled.

512K bytes of flash-simulated non-volatile configuration memory.
Base ethernet MAC Address       : FF:FF:FF:FF:FF:FF
Motherboard assembly number     : XX-XXXXX-XX
Power supply part number        : XXX-XXXX-XX
Motherboard serial number       : FOCXXXXXXXX
Power supply serial number      : FOCXXXXXXXX
Model revision number           : A0
Motherboard revision number     : A0
Model number                    : WS-C3560CX-8PC-S
System serial number            : FOCXXXXXXXX
Top Assembly Part Number        : XX-XXXX-XX
Top Assembly Revision Number    : A0
Version ID                      : V01
CLEI Code Number                : XXXXXXXXXX
Hardware Board Revision Number  : 0x02


Switch Ports Model                     SW Version            SW Image
------ ----- -----                     ----------            ----------
*    1 12    WS-C3560CX-8PC-S          15.2(4)E7             C3560CX-UNIVERSALK9-M


Configuration register is 0xF

3560CX#"""
    test_operations = [(channel_input_1, channel_output_1), (channel_input_2, channel_output_2)]
    conn = mocked_network_driver(test_operations)
    conn.default_desired_priv = "privilege_exec"
    conn.textfsm_platform = "not_real"
    results = conn.send_command(channel_input_2)
    parsed_results = results.textfsm_parse_output()
    assert isinstance(parsed_results, list)
    assert parsed_results == []



"""scrapli.channel.channel"""
import re
from logging import getLogger
from typing import List, Tuple

from scrapli.decorators import operation_timeout
from scrapli.helper import get_prompt_pattern, normalize_lines, strip_ansi
from scrapli.response import Response
from scrapli.transport.transport import Transport

LOG = getLogger("channel")


CHANNEL_ARGS = (
    "transport",
    "comms_prompt_pattern",
    "comms_return_char",
    "comms_ansi",
    "timeout_ops",
)


class Channel:
    def __init__(
        self,
        transport: Transport,
        comms_prompt_pattern: str = r"^[a-z0-9.\-@()/:]{1,32}[#>$]$",
        comms_return_char: str = "\n",
        comms_ansi: bool = False,
        timeout_ops: int = 10,
    ):
        """
        Channel Object

        Args:
            transport: Transport object of any transport provider (ssh2|paramiko|system|telnetlib)
                transport could in theory be any transport as long as it provides a read and a write
                method... obviously its probably always going to be scrapli transport though
            comms_prompt_pattern: raw string regex pattern -- use `^` and `$` for multi-line!
            comms_return_char: character to use to send returns to host
            comms_ansi: True/False strip comms_ansi characters from output
            timeout_ops: timeout in seconds for channel operations (reads/writes)

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.transport: Transport = transport
        self.comms_prompt_pattern = comms_prompt_pattern
        self.comms_return_char = comms_return_char
        self.comms_ansi = comms_ansi
        self.timeout_ops = timeout_ops

    def __str__(self) -> str:
        """
        Magic str method for Channel

        Args:
            N/A

        Returns:
            str: str for class object

        Raises:
            N/A

        """
        return "scrapli Channel Object"

    def __repr__(self) -> str:
        """
        Magic repr method for Channel

        Args:
            N/A

        Returns:
            str: repr for class object

        Raises:
            N/A

        """
        class_dict = self.__dict__.copy()
        class_dict.pop("transport")
        return f"scrapli Channel {class_dict}"

    def _restructure_output(self, output: bytes, strip_prompt: bool = False) -> bytes:
        """
        Clean up preceding empty lines, and strip prompt if desired

        Args:
            output: bytes from channel
            strip_prompt: bool True/False whether to strip prompt or not

        Returns:
            bytes: output of joined output lines optionally with prompt removed

        Raises:
            N/A

        """
        output = normalize_lines(output)

        if not strip_prompt:
            return output

        # could be compiled elsewhere, but allow for users to modify the prompt whenever they want
        prompt_pattern = get_prompt_pattern("", self.comms_prompt_pattern)
        output = re.sub(prompt_pattern, b"", output)
        return output

    def _read_chunk(self) -> bytes:
        """
        Private method to read chunk and strip comms_ansi if needed

        Args:
            N/A

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        new_output = self.transport.read()
        new_output = re.sub(b"\r", b"", new_output)
        if self.comms_ansi:
            new_output = strip_ansi(new_output)
        LOG.debug(f"Read: {repr(new_output)}")
        return new_output

    def _read_until_input(self, channel_input: bytes) -> bytes:
        """
        Read until all input has been entered.

        Args:
            channel_input: string to write to channel

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        output = b""
        while channel_input not in output:
            output += self._read_chunk()
        return output

    def _read_until_prompt(self, output: bytes = b"", prompt: str = "") -> bytes:
        """
        Read until expected prompt is seen.

        Args:
            output: bytes from previous reads if needed
            prompt: prompt to look for if not looking for base prompt (self.comms_prompt_pattern)

        Returns:
            bytes: output read from channel

        Raises:
            N/A

        """
        prompt_pattern = get_prompt_pattern(prompt, self.comms_prompt_pattern)

        while True:
            output += self._read_chunk()
            channel_match = re.search(prompt_pattern, output)
            if channel_match:
                return output

    @operation_timeout("timeout_ops", "Timed out determining prompt on device.")
    def get_prompt(self) -> str:
        """
        Get current channel prompt

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        prompt_pattern = get_prompt_pattern("", self.comms_prompt_pattern)
        self.transport.set_timeout(1000)
        self.transport.write(self.comms_return_char)
        LOG.debug(f"Write (sending return character): {repr(self.comms_return_char)}")
        output = b""
        while True:
            output += self._read_chunk()
            channel_match = re.search(prompt_pattern, output)
            if channel_match:
                self.transport.set_timeout()
                current_prompt = channel_match.group(0)
                return current_prompt.decode().strip()

    def send_input(self, channel_input: str, strip_prompt: bool = True) -> Response:
        """
        Primary entry point to send data to devices in shell mode; accept input and returns result

        Args:
            channel_input: string input to send to channel
            strip_prompt: strip prompt or not, defaults to True (yes, strip the prompt)

        Returns:
            Response: list of Response object(s)

        Raises:
            TypeError: if input is anything but a string

        """
        if not isinstance(channel_input, str):
            raise TypeError(
                f"`send_input` expects a single string, got {type(channel_input)}. "
                "to send a list of inputs use the `send_inputs` method instead"
            )
        response = Response(self.transport.host, channel_input)
        raw_result, processed_result = self._send_input(channel_input, strip_prompt)
        response.raw_result = raw_result.decode()
        response.record_response(processed_result.decode())
        return response

    def send_inputs(self, channel_inputs: List[str], strip_prompt: bool = True) -> List[Response]:
        """
        Primary entry point to send data to devices in shell mode; accept inputs and return results

        Args:
            channel_inputs: list of string inputs to send to channel
            strip_prompt: strip prompt or not, defaults to True (yes, strip the prompt)

        Returns:
            responses: list of Response object(s)

        Raises:
            TypeError: if anything but a list is passed for channel_inputs

        """
        if not isinstance(channel_inputs, list):
            raise TypeError(
                f"`send_inputs` expects a list of strings, got {type(channel_inputs)}. "
                "to send a single input use the `send_input` method instead"
            )

        responses = []
        for channel_input in channel_inputs:
            response = Response(self.transport.host, channel_input)
            raw_result, processed_result = self._send_input(channel_input, strip_prompt)
            response.raw_result = raw_result.decode()
            response.record_response(processed_result.decode())
            responses.append(response)
        return responses

    @operation_timeout("timeout_ops", "Timed out sending input to device.")
    def _send_input(self, channel_input: str, strip_prompt: bool) -> Tuple[bytes, bytes]:
        """
        Send input to device and return results

        Args:
            channel_input: string input to write to channel
            strip_prompt: bool True/False for whether or not to strip prompt

        Returns:
            result: output read from the channel

        Raises:
            N/A

        """
        bytes_channel_input = channel_input.encode()
        self.transport.session_lock.acquire()
        LOG.debug(f"Attempting to send input: {channel_input}; strip_prompt: {strip_prompt}")
        self.transport.write(channel_input)
        LOG.debug(f"Write: {repr(channel_input)}")
        self._read_until_input(bytes_channel_input)
        self._send_return()
        output = self._read_until_prompt()
        self.transport.session_lock.release()
        processed_output = self._restructure_output(output, strip_prompt=strip_prompt)
        # lstrip the return character out of the final result before storing, also remove any extra
        # whitespace to the right if any
        processed_output = processed_output.lstrip(self.comms_return_char.encode()).rstrip()
        return output, processed_output

    def send_inputs_interact(
        self, channel_inputs: List[str], hidden_response: bool = False
    ) -> Response:
        """
        Send inputs in an interactive fashion, used to handle prompts that occur after an input.

        Args:
            channel_inputs: list of four string elements representing...
                channel_input - initial input to send
                expected_prompt - prompt to expect after initial input
                response - response to prompt
                final_prompt - final prompt to expect
            hidden_response: True/False response is hidden (i.e. password input)

        Returns:
            Response: scrapli Response object

        Raises:
            TypeError: if inputs is not tuple or list

        """
        if not isinstance(channel_inputs, list):
            raise TypeError(f"`send_inputs_interact` expects a List, got {type(channel_inputs)}")
        channel_input, expectation, channel_response, finale = channel_inputs
        response = Response(
            self.transport.host,
            channel_input,
            expectation=expectation,
            channel_response=channel_response,
            finale=finale,
        )
        raw_result, processed_result = self._send_input_interact(
            channel_input, expectation, channel_response, finale, hidden_response
        )
        response.raw_result = raw_result.decode()
        response.record_response(processed_result.decode().strip())
        return response

    @operation_timeout("timeout_ops", "Timed out sending interactive input to device.")
    def _send_input_interact(
        self,
        channel_input: str,
        expectation: str,
        channel_response: str,
        finale: str,
        hidden_response: bool = False,
    ) -> Tuple[bytes, bytes]:
        """
        Respond to a single "staged" prompt and return results.

        Args:
            channel_input: string input to write to channel
            expectation: string of what to expect from channel
            channel_response: string what to respond to the `expectation`, or empty string to send
                return character only
            finale: string of prompt to look for to know when `done`
            hidden_response: True/False response is hidden (i.e. password input)

        Returns:
            output: output read from the channel

        Raises:
            N/A

        """
        bytes_channel_input = channel_input.encode()
        self.transport.session_lock.acquire()
        LOG.debug(
            f"Attempting to send input interact: {channel_input}; "
            f"\texpecting: {expectation};"
            f"\tresponding: {channel_response};"
            f"\twith a finale: {finale};"
            f"\thidden_response: {hidden_response}"
        )
        self.transport.write(channel_input)
        LOG.debug(f"Write: {repr(channel_input)}")
        self._read_until_input(bytes_channel_input)
        self._send_return()
        output = self._read_until_prompt(prompt=expectation)
        # if response is simply a return; add that so it shows in output likewise if response is
        # "hidden" (i.e. password input), add return, otherwise, skip
        if not channel_response or hidden_response is True:
            output += self.comms_return_char.encode()
        self.transport.write(channel_response)
        LOG.debug(f"Write: {repr(channel_response)}")
        self._send_return()
        LOG.debug(f"Write (sending return character): {repr(self.comms_return_char)}")
        output += self._read_until_prompt(prompt=finale)
        self.transport.session_lock.release()
        processed_output = self._restructure_output(output, strip_prompt=False)
        return output, processed_output

    def _send_return(self) -> None:
        """
        Send return char to device

        Args:
            N/A

        Returns:
            N/A  # noqa: DAR202

        Raises:
            N/A

        """
        self.transport.write(self.comms_return_char)
        LOG.debug(f"Write (sending return character): {repr(self.comms_return_char)}")
